# -*- coding: utf-8 -*-
"""
    gene
    ~~~~~~~~~~~~~~

    function for gene

    @Copyright: (c) 2018-04 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
import os

from bioutensil import basic


def read_gene_list(fn, sep='\t'):
    if not fn:
        return []
    gene_list = []
    for row in basic.safe_open(fn, 'r'):
        if row.startswith('#'):
            continue
        gene = row.strip().split(sep)[0]
        if gene:
            gene_list.append(gene)

    return gene_list

def read_gene_list_from_dir(in_dir):
    gene_set = set()
    for gs in read_gene_dict_from_dir(in_dir).values():
        gene_set = set(gs) | gene_set
    return list(gene_set)

def read_gene_dict_from_dir(in_dir):
    if not in_dir:
        return {}
    if os.path.isfile(in_dir):
        _, fn = os.path.split(in_dir)
        return {fn: read_gene_list(fn)}
    gene_dict = {}
    for fn in os.listdir(in_dir):
        in_fn = os.path.join(in_dir, fn)
        gene_dict[fn] = read_gene_list(in_fn)

    return gene_dict



import os
import pandas as pd
from pybedtools import BedTool


def read_chrom_size(assembly):
    fn_dir = os.path.dirname(os.path.abspath(__file__))
    fn = os.path.join(fn_dir, 'bio_database/{}.chrom.size'.format(assembly))
    df = pd.read_csv(fn, sep='\t')
    return df.set_index('chrom')['size'].to_dict()


def read_cytoband(assembly):
    fn_dir = os.path.dirname(os.path.abspath(__file__))
    fn = os.path.join(fn_dir, 'bio_database/{}.cytoBand.txt'.format(assembly))
    df = pd.read_csv(fn, sep='\t')
    df['cytoband'] = df['#chr'].str.replace('chr', '') + df['cytoband']
    df['startPos'] = df['startPos'] + 1
    df['region'] = df['#chr'].astype(str) + ':' + df['startPos'].astype(str) + '-' + df['endPos'].astype(str)
    return df.set_index('cytoband')[['region', 'color']].to_dict()




def get_cytoband_from_pos(cytoband_dict, chrom, pos):
    bin_list = ['{}\t{}'.format(v, k) for k, v in cytoband_dict['region'].items()]
    bed_str = '\n'.join(bin_list).replace(':', '\t').replace('-', '\t')
    bin_bed = BedTool(bed_str, from_string=True)

    pos_bed = BedTool('{}\t{}\t{}'.format(chrom, pos, pos+1), from_string=True)

    for i, hit in enumerate(pos_bed.window(bin_bed).overlap(cols=[2, 3, 5, 6])):
        return hit[6]

def get_cytoband_from_bin_str(cytoband_dict, bin_str):
    chrom = bin_str.split(':')[0]
    start, end = map(int, bin_str.split(':')[1].split('-'))
    pos = int((start+end)/2)
    return get_cytoband_from_pos(cytoband_dict, chrom, pos)



import os, sys, subprocess

def sort_files(fns, sorted_col, header=True):
    for fn in fns:
        sort_file(fn, sorted_col, header=header)
    return True

def sort_file(fn, sorted_col, header=True):
    path, name = os.path.split(fn)
    sorted_fn = os.path.join(path, name.replace('tsv','sortBy.%s.tsv' % str(sorted_col)))
    cmd = '(zcat %s | head -1; zcat %s | sed "1d" | sort ) | gzip -c > %s' % (fn, fn, sorted_fn)
    #print cmd
    subprocess.call(cmd, shell=True)
    #cmd = 'rm -f %s' % (fn)
    #subprocess.call(cmd, shell=True)
    return True




import math


import svgobject


class Axis(svgobject.Container):

    def __init__(self, insert, width, ticks_gap_width, 
            label='axis label', label_width=15, tick_self_width=3, height=10,
            mirror=False, h_margin=80, v_margin=10,
            **extra):
        self.ticks_gap_width = ticks_gap_width
        self.tick_self_width = tick_self_width

        self.tick_num = int(math.ceil(width/ticks_gap_width)) + 1
        self.length =  (self.tick_num - 1) * ticks_gap_width
        self.label = label
        self.label_width = label_width
        self.mirror = mirror

        size = self.length, height
        super(Axis, self).__init__(insert, size, 
                h_margin=h_margin, v_margin=v_margin, **extra)


    def render(self):
        related_insert = ('center', 'middle')
        insert = (0, 0)
        size = (self.length, 0)
        line = svgobject.Line(insert, size, stroke='black', parent=self._svg, related_insert=related_insert).render()
        # add axis label
        related_insert = ('center', 'right')
        label_text = svgobject.Text(self.label, self.insert, anchor_object=line, related_insert=related_insert).render()

        for i in range(self.tick_num):
            _, height = self.size

            insert = (i*self.ticks_gap_width, 0) # insert should be the up-left most
            if self.mirror:
                insert = (i*self.ticks_gap_width, -height/2.0)
            size = (self.label_width, height/2.0) # size should be positive

            tick_label = str(i)
            tick = Tick(insert, size, tick_label, 
                    anchor_object=line, self_pos='middle',
                    mirror=self.mirror).render()
            self._svg.add(tick)
   
        self._svg.add(line)
        self._svg.add(label_text)
        return self._svg 


class Tick(svgobject.Container):

    def __init__(self, insert, size, label, stroke='black', 
            rotate = 0, font_size=10, font_family='Arial',
            mirror = False, tick_self_width=3,
            **extra):

        super(Tick, self).__init__(insert, size, rotate=rotate, **extra)

        self.label = label
        self.stroke = stroke 
        self.font_size = font_size
        self.font_family = font_family
        self.mirror = mirror
        self.tick_self_width = tick_self_width

    def render(self):
        insert = 0, 0
        size = 0, self.tick_self_width 
        related_insert=('up','middle')
        if self.mirror:
            related_insert=('bottom','middle')

        tick = svgobject.Line(insert, size, stroke=self.stroke, transform=self.transform,
                parent=self._svg, related_insert=related_insert).render()

        related_insert = ('bottom', 'middle')
        if self.mirror:
            related_insert = ('up', 'middle')

        label_text = svgobject.Text(self.label, insert, anchor_object=tick, related_insert=related_insert, 
                font_size=self.font_size, font_family=self.font_family).render()

        self._svg.add(tick)
        self._svg.add(label_text)

        return self._svg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    svgfigure
    ~~~~~~~~~~~~~~~~~~~

    SVG Figure

    @Copyright: (c) 2018-09 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPDF, renderPM
import svgwrite

import svgbase
import text


class SVGFigure():

    def __init__(self,
                 out_prefix='demo',
                 format='svg',
                 title='SVG Figure',
                 size=(1000, 1000),
                 version='0.0', author='Lingxi Chen', email='chanlingxi@gmail.com', address="City University of Hong Kong",
                 info={'Info': 'Please tell something about this figure.'},
                 icon_related_insert='bottom:middle', info_related_insert='up:middle',
                 **kwargs
                 ):
        self.out_prefix = out_prefix
        self.format = format
        self.title = title
        self.size = size

        self.version = version
        self.author = author
        self.email = email
        self.address = address
        self.info = info
        self.info_related_insert = info_related_insert
        self.icon_related_insert = icon_related_insert

        self.dwg = svgwrite.Drawing(out_prefix + '.svg', size=size)
        self.canvas = svgbase.Container((0, 0), size).svgobj

    def draw(self):

        self._draw_info()
        self._draw_icon()

        self.dwg.add(self.canvas)
        self.dwg.save()
        self._convert()

    def _convert(self):
        if self.format == 'pdf':
            drawing = svg2rlg(self.out_prefix + '.svg')
            renderPDF.drawToFile(drawing, '{}.pdf'.format(self.out_prefix))

        if self.format == 'png':
            drawing = svg2rlg(self.out_prefix + '.svg')
            renderPM.drawToFile(drawing, "{}.png".format(self.out_prefix))

    def _draw_textarea(self, text_str, insert, parent, related_insert):
        textarea = text.TextArea(text_str, insert,
                                 parent=parent, related_insert=related_insert).render()
        return textarea

    def _draw_icon(self):
        icon_text = '\n'
        icon_text += '{}\n'.format(self.title)
        icon_text += 'v{}\n'.format(self.version)
        icon_text += 'Author: {}\n'.format(self.author)
        icon_text += 'Email: {}\n'.format(self.email)
        icon_text += self.address
        icon_text += '\n'
        icon = self._draw_textarea(icon_text, (0, 0),
                                   parent=self.canvas, related_insert=self.icon_related_insert)
        self.canvas.add(icon)

    def _draw_info(self):
        infos = ['', self.title]
        infos += [name + ': ' + value for name, value in self.info.items()]
        infos += ['']
        info_text = '\n'.join(infos)
        info = self._draw_textarea(info_text, (0, 0),
                                   parent=self.canvas, related_insert=self.info_related_insert)
        self.canvas.add(info)



# -*- coding: utf-8 -*-
"""
    converter
    ~~~~~~~~~~~~~~~~~~~

    converter
    SV <-> VCF Record

    @Copyright: (c) 2018-04 by Lingxi Chen (chanlingxi@gmail.com). 
    @License: LICENSE_NAME, see LICENSE for more details.
"""

import vcf_aux
import sv



if __name__ == "__main__":
    print(vcf2sv(1,3))
    print(vcf_aux.MyRecord())



import csv
import gzip
import sys

# csv.field_size_limit(sys.maxsize)

# --- load mapping function --- #
def str2int(x):
    try: x = int(x)
    except: x = 0
    return x

def str2float(x):
    try: x = float(x)
    except: x = 0.0
    return x

def str2list(sep=','):
    def func(x):
        return [] if x == '.' else x.split(sep)
    return func

# --- dump mapping function --- #
list2str = lambda l: ','.join([ str(x) for x in l]) if l else '.'

# --- reduce mapping function --- #
right = lambda x, y: y
left = lambda x, y: x
add = lambda x, y: x + y
list_append = lambda l, x: [x] if l == 'empty_list' else l + [x]

def assign_map(func2headers):
    _map = {}
    for func, headers in func2headers.items():
        for h in headers: _map[h] = func
    return _map

def assign_reduce_map(func_init2headers):
    _map = {}
    for func_init, headers in func_init2headers.items():
        func, init = func_init
        for h in headers: _map[(h,init)] = func
    return _map


class csvReader(object):

    def __init__(self, in_fn, load_map, headers=None, sep=',', reduce_map={}, has_header=False):
        self.in_fn = in_fn
        self.load_map = load_map
        self.headers = headers
        self.sep = sep
        self.reduce_map = reduce_map
        self.has_header = has_header

    def _load_row(self, row, filtered_funcs=[], required_headers='all'):
        error = []
        for header, filtered_func in filtered_funcs:
            if header in row:
                value = self.load_map.get(header, lambda x:x)(row[header])
                if not filtered_func(value):
                    error.append('%s=%s' % (header, str(value)))
        if error:
            return 'Error: '+','.join(error)

        if required_headers == 'all': required_headers = self.headers
        if not required_headers: required_headers = row.keys()

        data = {}
        for header in required_headers:
            value = row[header]
            value = self.load_map.get(header, lambda x: x)(value)
            data[header] = value
        return data

    def read_rows(self, filtered_funcs=[], required_headers='all'):

        if self.in_fn.endswith('.gz'):
            f = gzip.open(self.in_fn, 'rb')
        else:
            f = open(self.in_fn, 'rb')

        collection = csv.DictReader(f, fieldnames=self.headers, delimiter=self.sep, skipinitialspace=True)
        for i, row in enumerate(collection):
            if i == 0 and self.has_header: pass
            else:
                row = self._load_row(row, filtered_funcs, required_headers)
                if 'Error' not in row:
                    if self.has_header: i-=1
                    yield (i, row)

    def read_chunks(self, header, filtered_funcs=[], required_headers='all', reduce=False):
        chunks = []
        prev_value = None
        chunk_i = 0
        for i, row in self.read_rows(filtered_funcs, required_headers):
            value = row[header]
            if prev_value is None or prev_value == value: chunks.append(row)
            else: # update
                if i:
                    yield (prev_value, self.reduce_chunks(chunks, reduce))
                    chunk_i += 1
                chunks = [row]
            prev_value = value

        if chunks: yield (prev_value, self.reduce_chunks(chunks, reduce))

    def reduce_chunks(self, chunks, reduce):
        if not reduce: return chunks

        reduce_map = {  head_init:func for head_init, func in self.reduce_map.items() if head_init[0] in chunks[0].keys()}

        res = { header:init for header, init in reduce_map.keys() }
        for row in chunks:
            for header_init, func in reduce_map.items():
                header, _ = header_init
                res[header] = func(res[header], row[header])

        return [res]

class csvWriter():

    def __init__(self, out_fn, dump_map, headers, sep=','):
        self.out_fn = out_fn
        self.dump_map = dump_map
        self.headers = headers
        self.sep = sep

        self._write_header()

    def _write_header(self):
        if self.out_fn.endswith('.gz'):
            f = gzip.open(self.out_fn, 'wb')
        else:
            f = open(self.out_fn, 'wb')

        line = self.sep.join(self.headers) + '\n'
        f.write(line)
        return True

    def _dump_row(self, row):
        data = {}
        for header, value in row.items():
            value = self.dump_map.get(header, str)(value)
            data[header] = value
        return data

    def write_row(self, row, mod='ab'):


        if self.out_fn.endswith('.gz'):
            f = gzip.open(self.out_fn, mod)
        else:
            f = open(self.out_fn, mod)

        row = self._dump_row(row)
        # write
        line = self.sep.join([ row[header] for header in self.headers]) + '\n'
        f.write(line)
        f.close()
        return True

    def write_rows(self, rows, mod='ab'):

        if self.out_fn.endswith('.gz'):
            f = gzip.open(self.out_fn, mod)
        else:
            f = open(self.out_fn, mod)

        for row in rows:
            row = self._dump_row(row)
            # write
            line = self.sep.join([ row[header] for header in self.headers]) + '\n'
            f.write(line)

        f.close()
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    significant_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    @Copyright: (c) 2018-08 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
from scipy import stats
from skidmarks import wald_wolfowitz
from statsmodels.stats.multitest import fdrcorrection

import numpy as np
np.random.seed(12345678)

def fdr(pvals):
    return fdrcorrection(pvals, alpha=0.05, method='indep', is_sorted=False)

def exponent_distribution(x):
    '''
    null hypothesis: samples from exponent distribution
    '''
    if len(x) <= 2:
        return None, None
    param = stats.expon.fit(x)
    D, p = stats.kstest(x, 'expon', args=param)
    return p < 0.05, p


def multinomial_distribution(x):
    '''
    null hypothesis: samples from multinominal distribution
    '''
    # the expected frequencies are uniform and given by the mean of the observed frequencies.
    chisq, p = stats.chisquare(x)
    return p < 0.05, p


def wald_wolfowitz_test(x):
    '''
    null hypothesis: samples from alternating sequence
    010101010101

    '''
    if len(set(x)) == 1:
        return None, None, None
    if len(x) <= 2:
        return None, None, None
    p, z, n_runs, sd, mean = wald_wolfowitz(x).values()
    return p < 0.05, p, int(n_runs)/float(len(x))


def rank_sum_test(x1, x2):

    '''
    statistic : float
    The test statistic under the large-sample approximation that the rank sum statistic is normally distributed
    pvalue : float
    The two-sided p-value of the test
    '''
    s, p = stats.ranksums(x1, x2)

    return p < 0.05, p


def t_test(x1, x2, equal_var=True, ind=True):

    if ind:
        s, p = stats.ttest_ind(x1, x2, equal_var=equal_var)
    else:
        s, p = stats.ttest_rel(x1, x2)

    return p < 0.05, p


def anova_test(x1, x2):
    s, p = stats.f_oneway(x1, x2)
    return p < 0.05, p




from bioutensil.version import __version__
from bioutensil import bam
from bioutensil import bash
from bioutensil import basic
from bioutensil import clustering
# import cnv
# import complex_haplotype
from bioutensil import constants
# import converter
from bioutensil import csv_aux
from bioutensil import fasta
from bioutensil import gene
# from bioutensil import genome
from bioutensil import genomic_region
from bioutensil import module
from bioutensil import myio
from bioutensil import myprint
from bioutensil import psl_tran
from bioutensil import sequence
from bioutensil import shell
from bioutensil import structure_variant
# import vcf_aux

name = 'bioutensil'



# -*- coding: utf-8 -*-
"""
    cnv
    ~~~~~~~~~~~~~

    copy number variant

    @Copyright: (c) 2018-07 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""

from utils import myio


class CNVRecord(myio.Record):

    fields = 'id,chr,start,end,np,mean,arm,snvs,ai,median,Cn,mCn,fullCN,meanCn,purity'.split(',')  # tumor_percent

    def __init__(self, *args, **kwargs):
        super(CNVRecord, self).__init__(*args, **kwargs)

    def set(self,
            id=None,
            chr=None,
            start=None,
            end=None,
            np=None,
            mean=None,
            arm=None,
            snvs=None,
            ai=None,
            median=None,
            Cn=None,
            mCn=None,
            fullCN=None,
            meanCn=None,
            purity=None,
            **kwargs):

        self.id = self._validate(id, int)
        self.chr = self._validate(chr)
        self.start = self._validate(start, int)
        self.end = self._validate(end, int)
        self.np = self._validate(np, int)
        self.mean = self._validate(mean, float)
        self.arm = self._validate(arm)
        self.snvs = self._validate(snvs, int)
        self.ai = self._validate(ai, float)
        self.median = self._validate(median, float)
        self.Cn = self._validate(Cn, int)
        self.mCn = self._validate(mCn, int)
        self.fullCN = self._validate(fullCN)
        self.meanCn = self._validate(meanCn, float)
        self.purity = self._validate(purity, float)

    def parse(self, line):  # parse a stream_or_string to cnv object
        args = dict(zip(self.fields, line.strip().split(',')))
        self.set(**args)

    def __str__(self):
        return ','.join(str(getattr(self, field)) for field in self.fields)


class CNVReader(myio.Reader):

    def __init__(self, fn, cnv_cls=CNVRecord, sep=',', has_header=True):
        super(CNVReader, self).__init__(fn, record_cls=cnv_cls, sep=sep, has_header=has_header)


class CNVWriter(myio.Writer):

    def __init__(self, fn, cnv_cls=CNVRecord, sep=','):
        super(CNVWriter, self).__init__(fn, record_cls=cnv_cls, sep=sep)

'''
if __name__ == "__main__":
    fn = '/home/chenlingxi/mnt/chenlingxi/workspace/FS_Projects/WHOC.WGS_10X.batch01/CNACalling/patchwork/OC001/OC001/oc001t_Copynumbers.csv'

    for i, record in enumerate(CNVReader(fn)):
        meta, record = record
        print record.start, record.end, (record.end - record.start) / 1000
'''



from bplot import svgbase
from bplot import text


class Table(svgbase.Container):

    def __init__(self, insert, size, matrix,
                 rows=None,
                 cols=None,
                 show_value=None,
                 show_rows=None,
                 show_cols=None,
                 value_pos='middle:center',
                 show_grid=None,
                 font_size=10,
                 cell_fill='none',
                 value_fill='black',
                 value_cmap=None,
                 cell_cmap=None,
                 **kwargs
                 ):
        self.matrix = matrix
        self.rows = rows
        self.cols = cols
        self.show_value = show_value
        self.value_pos = value_pos
        self.show_rows = show_rows
        self.show_cols = show_cols

        self.show_grid = show_grid

        self.font_size = font_size

        self.value_fill = value_fill
        self.cell_fill = cell_fill
        self.value_cmap = value_cmap or []
        self.cell_cmap = cell_cmap or []

        self.ratio_x = ':'.join(['1']*len(self.matrix[0]))
        self.ratio_y = ':'.join(['1']*len(self.matrix))
        super(Table, self).__init__(insert, size,
                                    **kwargs)

        self._render()
        if rows:
            self._render_rows()
        if cols:
            self._render_cols()

    def _render(self):

        rows = self.split(split_ratio=(self.ratio_x, self.ratio_y))
        for i, row in enumerate(rows):
            for j, cell in enumerate(row):
                if self.show_grid:
                    cell_stroke = 'black'
                else:
                    cell_stroke = 'white'

                if self.cell_cmap:
                    cell_fill = self.cell_cmap[i][j]
                else:
                    cell_fill = self.cell_fill

                rect = svgbase.Rect((0, 0), cell.size,
                                    fill=cell_fill, stroke=cell_stroke)
                cell.svgobj.add(rect.svgobj)
                cell.rect = rect

                v = self.matrix[i][j]
                if self.value_cmap:
                    value_fill = self.value_cmap[i][j]
                else:
                    value_fill = self.value_fill
                if self.show_value:
                    self._set_label(cell, v,
                                    show_value=self.show_value,
                                    value_pos=self.value_pos,
                                    fill=value_fill)
                self.svgobj.add(cell.svgobj)

    def _render_rows(self):
        width = 10
        for row in self.rows:
            w = text.measure_text(''.join(str(x) for x in row), self.font_size) + 2*len(row)
            if w > width:
                width = w

        x, y = self.insert
        _, height = self.size
        insert = x - width, y
        size = width, height

        container = Table(insert, size, self.rows,
                          show_value=self.show_rows,
                          value_pos='iright:center',
                          font_size=self.font_size,
                          )
        self.trows = container

    def _render_cols(self):
        height = self.font_size*2

        x, y = self.insert
        width, _ = self.size
        insert = x, y - height
        size = width, height
        container = Table(insert, size, self.cols,
                          show_value=self.show_cols,
                          value_pos='middle:ibottom',
                          font_size=self.font_size,
                          )
        self.tcols = container

    def _set_label(self, cell, v,
                   show_value=None,
                   value_pos='',
                   fill='black',
                   ):
        try:
            v = round(v, 2)
        except Exception:
            pass
        if show_value == 'latex':
            name = v.replace('\\', '').replace('{', '').replace('}', '')
            size = cell.size[0]*2/3, cell.size[1]*2/3
            label = text.LatexText(v, name, (0, 0), size,
                                   font_size=self.font_size, fill='black',
                                   anchor_object=cell.rect, anchor_pos=value_pos,
                                   )
            cell.svgobj.add(label.svgobj)
            return

        if show_value == 'sign':
            if v > 0:
                v = '+'
            elif v < 0:
                v = '-'
            else:
                v = '0'

        label = svgbase.Text(v,
                             anchor_object=cell.rect,
                             anchor_pos=value_pos,
                             fill=fill,
                             font_size=self.font_size,
                             )
        cell.svgobj.add(label.svgobj)



# -*- coding: utf-8 -*-
"""
    sv.vcf
    ~~~~~~

    @Copyright: (c) 2018-03 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""

import gzip
import vcf
import copy

import structure_variant
import basic


class Record(vcf.model._Record, structure_variant.SVRecord):

    def __init__(self, *args, **kwargs):
        vcf.model._Record.__init__(self, *args, **kwargs)
        structure_variant.SVRecord.__init__(self)

    @property
    def to_sv(self):
        if self.var_type == 'INV':
            return '\n'.join(str(x) for x in self.bnd_records)
        else:
            return structure_variant.SVRecord.__str__(self)

    @property
    def var_type(self):
        if self.is_snp:
            return "snp"
        elif self.is_indel:
            return "indel"
        elif self.is_sv:
            return self.INFO.get('SVTYPE', 'sv')
        else:
            return "unknown"

    @property
    def secondary(self):
        return self.INFO.get('SECONDARY', False)

    @property
    def sv_end(self):
        return self.INFO.get('END', None)

    def set_sv_fields(self):
        pass

    @property
    def _get_general_sv_fields(self):
        args = {}
        args['id'] = self.ID
        args['qual'] = self.QUAL
        args['filter'] = self.FILTER
        args['span_reads'] = self.INFO.get('PE', [None])[0]
        args['junc_reads'] = self.INFO.get('SR', [None])[0]
        return args

    @property
    def _get_inv_bnd_records(self):  # transform INV format to BNDs format
        '''
        pos:            ab            cd
                        AC            GT
        +: 5' ----------oo------------oo---------- 3'
        -: 3' ----------oo------------oo---------- 5'

        INV Format:
        #CHROM  POS ID      REF ALT         QUAL    FILTER  INFO
        chr1    b   1       N   <INV>       10      PASS    SVTYPE=INV;END=c

        BNDs Format:
        - BDD1
        #CHROM  POS ID      REF ALT         QUAL    FILTER  INFO
        chr1    a   1.1_1   N   A]chr1:c]   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1
        chr1    c   1.1_2   N   G]chr1:a]   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1;SECONDARY

                        a
                        A
        +: 5' ----------o 3' ==================>
        -: 3' ----------o 5'                   |
                                               |
                                      c        |
                                      G        |
        +:            5' o------------o 3'     |
        -:            3' o------------o 5' <===

        - BND2
        chr1    b   1.2_1   N   [chr1:d[C   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1
        chr1    d   1.2_2   N   [chr1:b[T   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1:SECONDARY

                         b
                         C
        +:            5' o------------o 3'
        -:   <======= 3' o------------o 5'
             |
             |                         d
             |                         T
        +:   =====================> 5' o---------- 3'
        -:                          3' o---------- 5'

        '''
        args = {}
        args['chrom_5p'] = self.CHROM
        args['chrom_3p'] = self.CHROM
        args.update(self._get_general_sv_fields)
        args['meta_info'] = self._get_sv_meta_info   # no need to change for INV
        args['anno_info'] = self._get_sv_anno_info

        id = args['id']
        # BND 1
        args['id'] = id + '.1_1'
        args['bkpos_5p'] = self.POS - 1
        args['bkpos_3p'] = self.sv_end
        args['strand_5p'] = '+'     # ++ version
        args['strand_3p'] = '-'
        record1 = structure_variant.SVRecord(args=args)

        # BND2
        args['id'] = id + '.2_1'
        args['bkpos_5p'] = self.POS
        args['bkpos_3p'] = self.sv_end + 1
        args['strand_5p'] = '-'     # -- verson
        args['strand_3p'] = '+'
        record2 = structure_variant.SVRecord(args=args)
        return record1, record2

    @property
    def _get_inv_sv_fields(self):  # just for INV type comparision
        args = {}
        args['chrom_5p'] = self.CHROM
        args['bkpos_5p'] = self.POS
        args['strand_5p'] = '+'

        args['chrom_3p'] = self.CHROM
        args['bkpos_3p'] = self.sv_end
        args['strand_3p'] = '+'
        return args

    @property
    def _get_del_sv_fields(self):
        args = {}
        args['chrom_5p'] = self.CHROM
        args['bkpos_5p'] = self.POS
        args['strand_5p'] = '+'

        args['chrom_3p'] = self.CHROM
        args['bkpos_3p'] = self.sv_end
        args['strand_3p'] = '+'

        args['inner_ins'] = None
        return args

    @property
    def _get_dup_sv_fields(self):   # +: larger, small, switch start and end
        args = {}
        args['chrom_5p'] = self.CHROM
        args['bkpos_5p'] = self.sv_end
        args['strand_5p'] = '+'

        args['chrom_3p'] = self.CHROM
        args['bkpos_3p'] = self.POS
        args['strand_3p'] = '+'

        args['inner_ins'] = None
        return args

    @property
    def _get_unk_sv_fields(self):
        args = {}
        args['chrom_5p'] = self.CHROM
        args['bkpos_5p'] = self.POS
        args['strand_5p'] = '+'

        args['chrom_3p'] = self.CHROM
        args['bkpos_3p'] = self.sv_end
        args['strand_3p'] = '+'

        args['inner_ins'] = None
        return args

    @property
    def _get_bnd_sv_fields(self):
        '''
        VCF Format 1
        #CHROM  POS ID      REF ALT         QUAL    FILTER  INFO
        chr1    a   1_1     N   A]chr2:b]   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1
        chr2    b   1_2     N   [chr1:a[C   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1;SECONDARY

        SV Format
        #chrom_5p   bkpos_5p    strand_5p   chrom_3p    bkpos_3p    strand_3p
        chr1        a           +           chr2        b           +
                        a
                        A
        chr1 +: 5' ----------o 3' =========
        chr1 -: 3' ----------o 5'          |
                                           |
                                           |             b
                                           |             C
        chr2 +:                             ========> 5' o------------o 3'
        chr2 -:                                       3' o------------o 5'

        VCF Format 2
        #CHROM  POS ID      REF ALT         QUAL    FILTER  INFO
        chr1    a   1_1     N   A[chr2:b[   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1
        chr2    b   1_2     N   ]chr1:a]C   10      PASS    SVTYPE=BND;MATEID=.;EVENT=1;SECONDARY

        SV Format
        #chrom_5p   bkpos_5p    strand_5p   chrom_3p    bkpos_3p    strand_3p
        chr1        a           +           chr2        b           +
                                                         b
                                                         C
        chr2 +:                             ========> 5' o------------o 3'
        chr2 -:                            |          3' o------------o 5'
                        a                  |
                        A                  |
        chr1 +: 5' ----------o 3' =========
        chr1 -: 3' ----------o 5'

        '''
        args = {}
        alt = self.ALT[0]

        # if strand1 == '-' and strand2 == '-':   # inversion
        if alt.orientation and alt.remoteOrientation:   # inversion
            args['strand_5p'] = '-'
            args['strand_3p'] = '+'
        # elif strand1 == '+' and strand2 == '+':  # inversion
        elif not alt.orientation and not alt.remoteOrientation:  # inversion
            args['strand_5p'] = '+'
            args['strand_3p'] = '-'
        else:                                   # not inversion
            args['strand_5p'] = '+'
            args['strand_3p'] = '+'

        if alt.orientation:  # If the breakend is connected to sequence 3', extending to 5'
            args['chrom_5p'] = self.CHROM
            args['bkpos_5p'] = self.POS
            args['chrom_3p'] = alt.chr
            args['bkpos_3p'] = alt.pos

            args['inner_ins'] = alt.connectingSequence[:-1]

        else:               # If the breakend is connected to sequence 5', extending to 3'
            args['chrom_5p'] = alt.chr
            args['bkpos_5p'] = alt.pos
            args['chrom_3p'] = self.CHROM
            args['bkpos_3p'] = self.POS

            args['inner_ins'] = alt.connectingSequence[1:]

        return args

    @property
    def _get_sv_meta_info(self):
        return self.INFO

    @property
    def _get_sv_anno_info(self):
        self.anno_info = None


class LumpyRecord(Record):

    def __init__(self, *args, **kwargs):
        super(LumpyRecord, self).__init__(*args, **kwargs)

    def __str__(self):
        res = "Record(CHROM=%(CHROM)s, POS=%(POS)s, REF=%(REF)s, ALT=%(ALT)s), " % self.__dict__
        res += 'END={}, TYPE={}, STRANDS={})'.format(self.sv_end, self.var_type, self.strands)
        return res

    ''' newly add '''
    @property
    def strands(self):
        return self.INFO.get('STRANDS', None)

    def set_sv_fields(self):
        args = {}
        if self.var_type == 'INV':
            self.bnd_records = self._get_inv_bnd_records
            args = self._get_inv_sv_fields
            self.set(**args)
            return

        if self.var_type == 'DEL':
            args = self._get_del_sv_fields
        if self.var_type == 'DUP':
            args = self._get_dup_sv_fields
        if self.var_type == 'BND':
            args = self._get_bnd_sv_fields
        if self.var_type == 'DUP:TANDEM':
            args = self._get_dup_sv_fields
        if self.var_type == 'CNV':
            args = None         # ???

        args.update(self._get_general_sv_fields)

        args['meta_info'] = self._get_sv_meta_info
        args['anno_info'] = self._get_sv_anno_info

        self.set(**args)

    @property
    def _get_sv_meta_info(self):
        meta_info = copy.deepcopy(self.INFO)

        # delete tags
        del_tags = 'END,STRANDS,PE,SR'.split(',')
        for tag in del_tags:
            if tag in meta_info:
                del meta_info[tag]

        # switch tags
        switch_tags = [
            ('CIPOS', 'CIEND'),
            ('CIPOS95', 'CIEND95'),
            ('PRPOS', 'PREND')
        ]
        # longranger, only record CIPOS
        # If the breakend is connected to sequence 5', extending to 3', switch, lumpy didn't need to switch???
        if (self.var_type in ['DUP', 'DUP:TANDEM']) or (self.var_type == 'BND' and not self.ALT[0].orientation):
            for l_tag, r_tag in switch_tags:
                if l_tag in meta_info and r_tag in meta_info:
                    temp = meta_info[l_tag]
                    meta_info[l_tag] = meta_info[r_tag]
                    meta_info[r_tag] = temp

        # add new tags
        meta_info['GT'] = ','.join([sample['GT'] for sample in self.samples])
        meta_info['TOOL'] = 'Lumpy'

        # update mateid for INV
        if self.var_type == 'INV':
            meta_info['MATEID'] = None

        return meta_info


class LongrangerRecord(Record):

    def __init__(self, *args, **kwargs):
        super(LongrangerRecord, self).__init__(*args, **kwargs)

    def __str__(self):
        res = "Record(CHROM=%(CHROM)s, POS=%(POS)s, REF=%(REF)s, ALT=%(ALT)s), " % self.__dict__
        res += 'END={}, TYPE={})'.format(self.sv_end, self.var_type)
        return res

    @property
    def sv_type(self):
        if self.is_snp:
            return "snp"
        elif self.is_indel:
            return "indel"
        elif self.is_sv:
            return self.INFO.get('SVTYPE', 'sv')
        else:
            return "unknown"

    @property
    def sv_type2(self):
        return self.INFO.get('SVTYPE2', None)

    @property
    def var_type(self):
        if self.sv_type == 'BND' and not self.sv_type2:
            return self.sv_type2
        else:
            return self.sv_type

    def set_sv_fields(self):
        args = {}
        if self.var_type == 'INV':
            self.bnd_records = self._get_inv_bnd_records
            args = self._get_inv_sv_fields
            self.set(**args)
            return

        if self.var_type == 'DEL':
            args = self._get_del_sv_fields
        if self.var_type == 'DUP':
            args = self._get_dup_sv_fields
        if self.var_type == 'BND':
            args = self._get_bnd_sv_fields
        if self.var_type == 'DUP:TANDEM':
            args = self._get_dup_sv_fields

        if self.var_type == 'UNK':
            args = self._get_unk_sv_fields  # ???

        args.update(self._get_general_sv_fields)

        args['meta_info'] = self._get_sv_meta_info
        args['anno_info'] = self._get_sv_anno_info

        self.set(**args)

    @property
    def _get_sv_meta_info(self):
        meta_info = copy.deepcopy(self.INFO)

        # switch tags
        switch_tags = [
            ('CIPOS', 'CIEND'),
        ]
        if self.var_type in ['DUP', 'DUP:TANDEM']:
            for l_tag, r_tag in switch_tags:
                temp = meta_info[l_tag]
                meta_info[l_tag] = meta_info[r_tag]
                meta_info[r_tag] = temp

        # add new tags
        meta_info['GT'] = ','.join([sample['GT'] for sample in self.samples])
        meta_info['TOOL'] = 'Longranger'

        return meta_info


class Reader(vcf.parser.Reader):

    def __init__(self, record_cls=Record, *args, **kwargs):
        self._record_cls = record_cls
        super(Reader, self).__init__(*args, **kwargs)

    def next(self):
        '''Return the next record in the file.'''
        line = next(self.reader)
        row = self._row_pattern.split(line.rstrip())
        chrom = row[0]
        if self._prepend_chr:
            chrom = 'chr' + chrom
        pos = int(row[1])

        if row[2] != '.':
            ID = row[2]
        else:
            ID = None

        ref = row[3]
        alt = self._map(self._parse_alt, row[4].split(','))

        try:
            qual = int(row[5])
        except ValueError:
            try:
                qual = float(row[5])
            except ValueError:
                qual = None

        filt = self._parse_filter(row[6])
        info = self._parse_info(row[7])

        try:
            fmt = row[8]
        except IndexError:
            fmt = None
        else:
            if fmt == '.':
                fmt = None

        record = self._record_cls(chrom, pos, ID, ref, alt, qual, filt, info, fmt, self._sample_indexes)  # change the _Record type to Record

        if fmt is not None:
            samples = self._parse_samples(row[9:], fmt, record)
            record.samples = samples

        return record

    def _parse_filter(self, filt_str):
        '''Parse the FILTER field of a VCF entry into a Python list

        NOTE: this method has a cython equivalent and care must be taken
        to keep the two methods equivalent
        '''
        if filt_str == '.':
            return None
        elif filt_str == 'PASS':
            return []
        else:
            return filt_str.split(';')


class VCF(object):

    def __init__(self, vcf_fn, allowance=10, record_cls=Record):
        self._vcf_fn = vcf_fn
        self._allowance = allowance
        self._record_cls = record_cls

    @property
    def reader(self):
        # return Reader(basic.safe_open(self._vcf_fn, 'r'))
        compressed = self._vcf_fn.endswith('gz')
        return Reader(filename=self._vcf_fn, compressed=compressed, record_cls=self._record_cls)

    def read(self, ex_filter_funcs=[], in_filter_funcs=[]):
        for record in self.reader:
            if (not any(filter_func(record) for filter_func in ex_filter_funcs)) \
                    and (all(filter_func(record) for filter_func in in_filter_funcs)):
                yield record

    @property
    def to_sv(self):  # vcf to sv string
        for record in self.reader:
            record.set_sv_fields()
            yield 'whole', record.to_sv

    @property
    def set_sv_fields(self):  # vcf to sv record
        for record in self.reader:
            record.set_sv_fields()
            yield 'whole', record

    def write(self, out_fn, records, fmt='vcf'):
        if fmt == 'vcf':
            writer = vcf.Writer(open(out_fn, 'wb'), self.reader)
        if fmt == 'sv':
            writer = structure_variant.Writer(out_fn)
        for record in records:
            writer.write_record(record)

    def write_inter_diff(self, shared_fn, diff_fn, shared_or_diff, fmt='vcf'):
        if fmt == 'vcf':
            inter_writer = vcf.Writer(open(shared_fn, 'wb'), self.reader)
            diff_writer = vcf.Writer(open(diff_fn, 'wb'), self.reader)
        if fmt == 'sv':
            inter_writer = structure_variant.Writer(shared_fn)
            diff_writer = structure_variant.Writer(diff_fn)
        for tag, record in shared_or_diff:
            if tag == 'intersection':
                inter_writer.write_records([record])
            else:
                diff_writer.write_records([record])

    def fetch(self, chrom, start, end, fmt='vcf', ex_filter_funcs=[], in_filter_funcs=[]):  # check
        try:
            records = self.reader.fetch(chrom, start, end)
        except:
            return

        for record in records:
            if (not any(filter_func(record) for filter_func in ex_filter_funcs)) \
                    and (all(filter_func(record) for filter_func in in_filter_funcs)):

                if fmt == 'sv':
                    record.set_sv_fields()
                yield record

    def inter_diff(self, other, ex_filter_funcs=[], in_filter_funcs=[]):
        for i, self_record in enumerate(self.read(ex_filter_funcs=ex_filter_funcs, in_filter_funcs=in_filter_funcs)):
            self_record.set_sv_fields()
            chrom = self_record.chrom_5p
            pos = self_record.bkpos_5p
            start, end = pos - self._allowance, pos + self._allowance

            other_records = other.fetch(chrom, start, end, fmt='sv', ex_filter_funcs=ex_filter_funcs, in_filter_funcs=in_filter_funcs)
            if self_record in other_records:  # variants in self that also in other, self record is written
                yield 'intersection', self_record
            else:  # variants in self but not in other
                yield 'difference', self_record

    def intersection(self, other):
        for tag, record in self.inter_diff(other):
            if tag == 'intersection':
                yield self_record

    def difference(self, other):
        for tag, record in self.inter_diff(other):
            if tag == 'difference':
                yield self_record



import svgwrite

import svgobject

class ColorBar(svgobject.Container):

    def __init__(self, insert, size, name, min_value, max_value, color, font_size=10, **extra):

        self.name = name
        self.min_value = min_value
        self.max_value = max_value
        self.color = color
        self.font_size = font_size

        super(ColorBar, self).__init__(insert, size, **extra)

    def render(self):

        size = (50,20)
        related_insert = ('center', 'middle')
        defs, colorbar = svgobject.LinearGradientRect(self.insert, size, self.color, 0.3, self.color, 1.0, parent=self._svg, related_insert=related_insert).render()
        self._svg.add(colorbar)
        self._svg.defs.add(defs)

        related_insert = ('up', 'middle')
        name_text = svgobject.Text(self.name, self.insert, font_size=self.font_size, anchor_object=colorbar, related_insert=related_insert).render()
        self._svg.add(name_text)

        min_value_str = str(self.min_value)
        related_insert = ('bottom', 'left')
        min_value_text = svgobject.Text(min_value_str, self.insert, font_size=self.font_size, anchor_object=colorbar, related_insert=related_insert).render()
        self._svg.add(min_value_text)


        max_value_str = str(round(self.max_value, 4))
        related_insert = ('bottom', 'right')
        max_value_text = svgobject.Text(max_value_str, self.insert, font_size=self.font_size, anchor_object=colorbar, related_insert=related_insert).render()
        self._svg.add(max_value_text)

        return self._svg









import svgobject
import shape

class Chrom(svgobject.Container):

    def __init__(self, insert, size,
            chrom, p_start, p_end, q_start, q_end,
            chrom_width=3, rx=5, ry=5,
            font_size=10, font_family='Arial',
            h_margin=80, v_margin=5, cytoband=None,
            opacity=0.8,
            **extra):


        self.chrom = chrom
        self.p_start = p_start
        self.p_end = p_end
        self.q_start = q_start
        self.q_end = q_end

        self.cytoband_info=cytoband
        self.cytoband_color_scheme={
                'stalk': '#f0f0f0',
                'gvar': '#f0f0f0',
                'acen': '#960015', #'#f0f0f0',
                'gneg': '#d9d9d9',
                'gpos25': '#bdbdbd',
                'gpos50': '#969696',
                'gpos75': '#737373',
                'gpos100': '#525252',
        }

        self.chrom_width = chrom_width
        self.rx = rx
        self.ry = ry
        self.font_size = font_size
        self.font_family = font_family
        self.h_margin=h_margin
        self.v_margin=v_margin

        size = (q_end, chrom_width)
        super(Chrom, self).__init__(insert, size, h_margin=h_margin, v_margin=v_margin, **extra)


    def render(self):
        anchor = self._render_cytoband()
        self._render_label(anchor)
        return self._svg

    def _render_label(self, anchor):
        # label
        label_insert = (None, None)
        related_insert = ('center', 'left')
        label = svgobject.Text(self.chrom, label_insert, font_size=self.font_size, font_family=self.font_family, anchor_object=anchor, related_insert=related_insert).render()
        self._svg.add(label)

    def _render_arm(self):
        # p_arm
        p_insert = (self.x + self.h_margin + self.p_start, 0)
        p_size = (self.p_end - self.p_start, self.chrom_width)
        q_insert = (self.x + self.h_margin + self.q_start, 0)
        q_size = (self.q_end - self.q_start, self.chrom_width)
        for arm_name, insert, size, arm_start in [('q', q_insert, q_size, self.q_start), ('p', p_insert, p_size, self.p_start)]:
            arm = Arm(insert, size, self.chrom, arm_name, arm_start,
                    cytoband=self.cytoband_info,
                    rx=self.rx, ry=self.ry,
                    parent=self._svg, related_insert='center').render()
            self._svg.add(arm)
        return arm

    def _render_cytoband(self):
        insert = self.x + self.h_margin, 0
        size = self.q_end - self.p_start, self.chrom_width
        cytoband = CytoBand(insert, size, self.chrom,
                opacity=self.opacity,
                cytoband=self.cytoband_info, cytoband_color_scheme=self.cytoband_color_scheme,
                parent=self._svg, related_insert='center').render()
        self._svg.add(cytoband)

        return cytoband

class CytoBand(svgobject.Container):
    def __init__(self, insert, size,
            chrom, cytoband=None, cytoband_color_scheme=None,
            **extra):

        super(CytoBand, self).__init__(insert, size, **extra)
        self.chrom = chrom

        self.cytoband_info=cytoband
        if not cytoband_color_scheme:
            cytoband_color_scheme= {
                'stalk': '#f0f0f0',
                'gvar': '#f0f0f0',
                'acen': '#f0f0f0',
                'gneg': '#d9d9d9',
                'gpos25': '#bdbdbd',
                'gpos50': '#969696',
                'gpos75': '#737373',
                'gpos100': '#525252',
            }
        self.cytoband_color_scheme = cytoband_color_scheme



    def render(self):
        width, height = self.size
        cytoband_info = self.cytoband_info['cytoband']
        chrom2endposition = self.cytoband_info['chrom2endposition']
        for chrom, position, start, end, color  in cytoband_info:
            if chrom != self.chrom: continue
            if color == 'acen':
                if 'p' in position:
                    points = [(start, 0), (start, height), (end, 2.0*height/3), (end, 1.0*height/3)]
                if 'q' in position:
                    points = [(start, 1.0*height/3), (start, 2.0*height/3), (end, height), (end, 0)]
                stroke_pattern = [0,1,0,1]
                cytoband = shape.Polygon(points,
                        stroke_width=0.2, stroke_pattern=stroke_pattern,
                        opacity=self.opacity,
                        fill=self.cytoband_color_scheme[color],
                        ).render()
            else:
                insert = start, 0
                size = end - start, height
                front, end = chrom2endposition[chrom]
                if position == front:   stroke_lines = ['up','bottom','left']
                elif position == end:   stroke_lines = ['up','bottom','right']
                else:                   stroke_lines = ['up','bottom']
                cytoband = svgobject.Rect(insert, size,
                        stroke_width=0.3, stroke_lines=stroke_lines,
                        opacity=self.opacity,
                        fill=self.cytoband_color_scheme[color],
                        ).render()
            self._svg.add(cytoband)
        return self._svg


class Arm(svgobject.Container):
    def __init__(self, insert, size,
            chrom, name='p', arm_start=0,
            cytoband=None,
            rx=5, ry=5, opacity=0.5,
            **extra):

        super(Arm, self).__init__(insert, size, **extra)
        self.chrom = chrom
        self.name = name
        self.arm_start = arm_start

        self.cytoband_info=cytoband
        self.cytoband_color_scheme={
                'stalk': '#f0f0f0',
                'gvar': '#f0f0f0',
                'acen': '#f0f0f0',
                'gneg': '#d9d9d9',
                'gpos25': '#bdbdbd',
                'gpos50': '#969696',
                'gpos75': '#737373',
                'gpos100': '#525252',
        }

        self.rx = rx
        self.ry = ry
        self.opacity = opacity


    def render(self):
        self._render_cytoband()
        insert = (0,0)
        self.arm = svgobject.Rect(insert, self.size,
                rx=self.rx, ry=self.ry, opacity=self.opacity).render()
        self._svg.add(self.arm)
        return self._svg

    def _render_cytoband(self):
        width, height = self.size
        for chrom, position, start, end, color  in self.cytoband_info:
            if chrom != self.chrom: continue
            if self.name not in position: continue
            insert = start - self.arm_start , 0
            size = end - start, height
            rx = 0
            ry = 0
            cytoband = svgobject.Rect(insert, size, rx=rx, ry=ry,
                    fill=self.cytoband_color_scheme[color], stroke='white',
                    parent=self._svg, related_insert='center').render()
            self._svg.add(cytoband)






# -*- coding: utf-8 -*-
"""
    utils.bam
    ~~~~~~~~~~~~~~~~~

    Function to process BAM file     

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com). 
    @License: LICENSE_NAME, see LICENSE for more details.
"""


# check if hex_x is in dec_flag
def in_flag(hex_x, dec_flag):
    # convert hex_x into decimal 
    dec_x = int(hex_x, 16)
    # operation
    dec_res = dec_x & dec_flag

    return dec_res == dec_x

'''
print in_flag('0x1',3)
print in_flag('0x10',81), bin(81)
print in_flag('0x100',81), bin(81)
print in_flag('0x10',82), bin(82)
'''





# -*- cod:  -*-
"""
    utils.sv
    ~~~~~~~~

    Structure variation definition.

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""


class Deletion(object):

    """Docstring for Deletion. """

    def __init__(self, l_bp, r_bp, seq='', supports=0):
        self.l_bp = l_bp  # left breakpoint
        self.r_bp = r_bp  # right breakpoint

        self.seq = seq      # seq deleted
        self.supports = supports    # support read count

    def __str__(self):
        pass


class Insertion(object):

    """Docstring for Insertion. """

    def __init__(self, seq, l_bp, r_bp, supports=0):
        self.seq = seq  # seq inserted
        self.l_bp = l_bp    # left breakpoint
        self.r_bp = r_bp    # right breakpoint
        self.supports = supports    # support read count

    def __str__(self):
        pass


class TandemDuplication(object):

    """Docstring for TandemDuplication. """

    def __init__(self, seq, l_bp, r_bp, supports=0):
        self.seq = seq  # seq duplicated
        self.l_bp = l_bp    # left breakpoint
        self.r_bp = r_bp    # right breakpoint
        # self.dup_num = dup_num  # seq duplication times
        self.supports = supports    # support read count

    def __str__(self):
        pass


class Inversion(object):

    """Docstring for Inversion. """

    def __init__(self):
        """TODO: to be defined1. """
        pass


class ITX(object):

    """Docstring for ITX. """

    def __init__(self):
        """TODO: to be defined1. """


class CTX(object):

    """Docstring for CTX. """

    def __init__(self):
        """TODO: to be defined1. """



import svgwrite


class SVGObject(object):
    def __init__(self, insert, size,
                 fill='white', opacity=1,
                 stroke='black', stroke_width=1, stroke_dasharray=[], stroke_lines=['outline'],
                 parent=None,
                 anchor_object=None,
                 related_insert='',
                 self_pos='',
                 rotate=None,
                 transform=None,
                 profile='full',
                 **kwargs
                 ):

        self._dwg = svgwrite.Drawing(profile=profile)

        self.insert = insert
        self.x, self.y = insert
        self.size = size

        self.fill = fill
        self.opacity = opacity

        self.stroke = stroke
        self.stroke_width = stroke_width
        self._stroke_lines = stroke_lines
        if stroke_dasharray:
            self.stroke_dasharray = stroke_dasharray
        elif stroke_lines and self.size:
            self._draw_stroke_lines()

        # text
        self.dx = None
        self.dy = None
        self.style = None

        # line

        self.rotate = rotate
        self.transform = transform
        if rotate is not None:
            self._update_rotate()

        # the last part
        self.parent = parent
        self.anchor_object = anchor_object
        self.related_insert = related_insert
        self.self_pos = self_pos

        self._update_insert()
        self._update_pos()

    def _update_insert(self):
        # use anchor related position

        if self.anchor_object:  # absolute position
            if isinstance(self.anchor_object, svgwrite.shapes.Rect):
                x, y = self.anchor_object['x'], self.anchor_object['y']
                width, height = self.anchor_object['width'], self.anchor_object['height']
            if isinstance(self.anchor_object, svgwrite.shapes.Line):
                # line object
                x, y = self.anchor_object['x1'], self.anchor_object['y1']
                x2, y2 = self.anchor_object['x2'], self.anchor_object['y2']
                width, height = x2 - x, y2 - y
            if isinstance(self.anchor_object, svgwrite.shapes.Circle):
                r = self.anchor_object['r']
                height, width = r * 2, r * 2
                x, y = self.anchor_object['cx'] - r, self.anchor_object['cy'] - r

        if self.parent: # related position based on parent
            x, y = 0, 0
            width, height = self.parent['width'], self.parent['height']

        if 'up' in self.related_insert:             self.y = y
        elif 'center' in self.related_insert:       self.y = y + height/2.0
        elif 'bottom' in self.related_insert:       self.y = y + height
        elif self.anchor_object:                self.y += y

        if 'left' in self.related_insert:           self.x = x
        elif 'middle' in self.related_insert:       self.x = x + width/2.0
        elif 'right' in self.related_insert:        self.x = x + width
        elif self.anchor_object:                self.x += x

        self.insert = (self.x, self.y)

    def _update_pos(self):
        if self.self_pos:
            self._update_pos_by_self_pos()
        else:
            pass
            # self._update_pos_by_related_insert()

    def _update_pos_by_related_insert(self):
        width, height = self.size
        if 'middle' in self.related_insert:
            self.x -= width/2.0
        if 'right' in self.related_insert:
            self.x -= width
        if 'center' in self.related_insert:
            self.y -= height/2.0
        if 'bottom' in self.related_insert:
            self.y -= height

        self.insert = (self.x, self.y)

    def _update_pos_by_self_pos(self):
        width, height = self.size
        if 'right' in self.self_pos:      self.x
        if 'middle' in self.self_pos:     self.x -= width/2.0
        if 'left' in self.self_pos:      self.x -= width
        if 'bottom' in self.self_pos:     self.y
        if 'center' in self.self_pos:     self.y -= height/2.0
        if 'up' in self.self_pos:     self.y -= height

        self.insert = (self.x, self.y)

    def _update_rotate(self):
        # horizontal as default
        if self.rotate is not None:
            transform = 'rotate({} {} {})'.format(self.rotate, self.x, self.y)
            self.transform = transform

    def _draw_stroke_lines(self):
        width, height = self.size
        if 'outline' in self._stroke_lines: self._stroke_lines += ['up','right','bottom','left']
        pattern = [int(x in self._stroke_lines) for x in ['up','right','bottom','left']]
        size = [width, height, width, height]
        stroke_dasharray = []
        prev_pattern = -1
        element = 0
        for pattern, size in zip(pattern, size):
            if prev_pattern == pattern or prev_pattern == -1:
                element += size
            else: # new
                # update prev
                stroke_dasharray.append(element)
                element = size
            prev_pattern = pattern


        stroke_dasharray.append(element)

        if 'up' not in self._stroke_lines:
            stroke_dasharray = [0] + stroke_dasharray

        self.stroke_dasharray=','.join(map(str,stroke_dasharray))


class Rect(SVGObject):
    def __init__(self, insert, size,
            rx=0, ry=0, fill='white', opacity=1,
            **extra):
        super(Rect, self).__init__(insert, size, fill=fill, opacity=opacity, **extra)
        self.rx = rx
        self.ry = ry


    def render(self):
        rect = self._dwg.rect(self.insert, self.size,
                rx=self.rx, ry=self.ry,
                stroke=self.stroke, stroke_width=self.stroke_width, stroke_dasharray=self.stroke_dasharray,
                fill=self.fill, opacity=self.opacity)
        return rect


class Circle(SVGObject):
    def __init__(self, insert, size,
                 fill='white', opacity=1,
                 **extra):
        super(Circle, self).__init__(insert, size, fill=fill, opacity=opacity, **extra)
        self.r = size[0]/2.0

    def render(self):
        center = self.insert[0] + self.r, self.insert[1] + self.r
        rect = self._dwg.circle(center, self.r,
                                stroke=self.stroke, stroke_width=self.stroke_width, stroke_dasharray=self.stroke_dasharray,
                                fill=self.fill, opacity=self.opacity)
        return rect


class Line(SVGObject):
    def __init__(self, insert, size, **extra):
        super(Line, self).__init__(insert, size, **extra)

    def render(self):
        width, height = self.size
        end = self.x + width, self.y + height
        return self._dwg.line(self.insert, end, stroke=self.stroke)


class Text(SVGObject):

    def __init__(self, text, insert=(0, 0),
                 font_size=10, font_family='Arial', font_style='normal',
                 fill='black', letter_spacing=0,
                 **extra):
        super(Text, self).__init__(insert, None, **extra)
        self.text = text
        self.font_size = font_size
        self.font_family = font_family
        self.font_style = font_style
        self.fill = fill
        self.letter_spacing = letter_spacing

    def _update_pos(self):
        if 'up' in self.related_insert:
            self.dy = ['-0.3em']
        if 'center' in self.related_insert:
            self.dy = ['0.3em']
        if 'bottom' in self.related_insert:
            self.dy = ['1em']

        if 'left' in self.related_insert:
            self.style = 'text-anchor: end'
        if 'right' in self.related_insert:
            self.style = 'text-anchor: start'
        if 'middle' in self.related_insert:
            self.style = 'text-anchor: middle'
        if 'center' not in self.related_insert:
            self.style = 'text-anchor: middle'

        if self.style == 'text-anchor: end':
            self.dx = ['-0.5em']
        if self.style == 'text-anchor: start':
            self.dx = ['0.5em']

    def render(self):
        return self._dwg.text(self.text, self.insert, dx=self.dx, dy=self.dy,
                              style=self.style,
                              fill=self.fill,
                              font_size=self.font_size,
                              font_family=self.font_family,
                              font_style=self.font_style,
                              letter_spacing=self.letter_spacing)




class LinearGradientRect(Rect):

    def __init__(self, insert, size, start_color, start_opacity, end_color, end_opacity, **extra):
        super(LinearGradientRect, self).__init__(insert, size, **extra)

        self._start_color = start_color
        self._start_opacity = start_opacity
        self._end_color = end_color
        self._end_opacity = end_opacity

    def render(self):
        horizontal_gradient = self._dwg.linearGradient((0, 0), (1, 0))
        horizontal_gradient.add_stop_color(0, self._start_color, self._start_opacity)
        horizontal_gradient.add_stop_color(1, self._end_color, self._end_opacity)

        rect = self._dwg.rect(self.insert, self.size, fill=horizontal_gradient.get_paint_server(default='currentColor'))

        return horizontal_gradient, rect


class Container(SVGObject):
    def __init__(self, insert=(0,0), size=(10,10), stroke='gray',
            h_margin=0, v_margin=0,
            **extra):
        self.stroke = stroke

        self.h_margin = h_margin
        self.v_margin = v_margin
        width, height = size
        size = width+2*h_margin, height+2*v_margin

        super(Container, self).__init__(insert, size, **extra)
        self._svg = svgwrite.container.SVG(self.insert, self.size) # update the data


        #self.draw_lines('right','left','up','bottom')

    def split(self, split_pattern=None, split_ratio=None):
        width, height = self.size
        if split_pattern:
            split_x, split_y = split_pattern
            split_ratio_x = [1.0/split_x] * split_x
            split_ratio_y = [1.0/split_y] * split_y
        if split_ratio:
            split_ratio_x, split_ratio_y = split_ratio
            split_ratio_x = list(map(int, split_ratio_x.split(':')))
            split_ratio_y = list(map(int, split_ratio_y.split(':')))
            split_ratio_x = list(map(lambda n: 1.0*n/sum(split_ratio_x), split_ratio_x))
            split_ratio_y = list(map(lambda n: 1.0*n/sum(split_ratio_y), split_ratio_y))

        items = []
        child_y = 0
        for j in range(len(split_ratio_y)):
            child_height = height*split_ratio_y[j]

            line = []
            child_x = 0
            for i in range(len(split_ratio_x)):
                child_width = width*split_ratio_x[i]

                insert = child_x, child_y
                size = child_width, child_height
                child = Container(insert, size)
                line.append(child)

                child_x += child_width

            items.append(line)
            child_y += child_height
        return items

    def draw_lines(self, *names):
        names = list(names)
        if 'outline' in names:
            names += ['left', 'right', 'up', 'bottom']
        width, height = self.size
        if 'left' in names:
            line = Line((0,0), (0,height), parent=self._svg, stroke=self.stroke).render()
            self._svg.add(line)
        if 'right' in names:
            line = Line((width,0), (0,height), parent=self._svg, stroke=self.stroke).render()
            self._svg.add(line)
        if 'middle' in names:
            line = Line((width/2,0), (0,height), parent=self._svg, stroke=self.stroke).render()
            self._svg.add(line)

        if 'up' in names:
            line = Line((0,0), (width,0), parent=self._svg, stroke=self.stroke).render()
            self._svg.add(line)
        if 'bottom' in names:
            line = Line((0,height), (width,0), parent=self._svg, stroke=self.stroke).render()
            self._svg.add(line)
        if 'center' in names:
            line = Line((0,height/2), (width,0), parent=self._svg, stroke=self.stroke).render()
            self._svg.add(line)

    def render(self):
        return self._svg







# -*- coding: utf-8 -*-
"""
    module
    ~~~~~~~~~~~~~~~~

    @Copyright: (c) 2018-04 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
import os
from itertools import tee


class Module(object):

    def __init__(self, in_data=None, in_fn=None,
                 reader_cls=None, file_type=None,
                 record_cls=None,
                 write=False, write_stats=True, write_headers=False,
                 write_fields=True, dynamic_fields=False,
                 writer_cls=None, write_mode='whole', out_dir='',
                 prefix='', suffix='', chunk_size=1000,
                 sample=None,
                 **extra):
        self.sample = sample

        self.in_fn = in_fn
        self.in_data = in_data
        self.reader_cls = reader_cls
        self.file_type = file_type
        self.record_cls = record_cls

        if in_data:
            if file_type:
                self.in_iter = self.reader_cls(in_data=in_data, sample=sample, file_type=file_type)
            else:
                self.in_iter = self.reader_cls(in_data=in_data, sample=sample)
        elif in_fn:
            if file_type:
                self.in_iter = self.reader_cls(in_fn=in_fn, sample=sample, file_type=file_type)
            else:
                self.in_iter = self.reader_cls(in_fn=in_fn, sample=sample)
        else:
            self.in_iter = in_data

        self.write = write
        self.write_stats = write_stats
        self.write_headers = write_headers
        self.write_fields = write_fields
        self.dynamic_fields = dynamic_fields
        self.writer_cls = writer_cls
        self.out_fn = None
        self.out_fns = []

        self.write_mode = write_mode
        self.out_dir = out_dir
        self.prefix = prefix
        self.suffix = suffix
        self.chunk_size = chunk_size

        self.meta_dict = {}
        self.writer_dict = {}
        self.meta_count = {}

    def evaluate(self):
        gen1, gen2 = tee(self)

        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)
        for meta, record in gen1:
            if self.write:
                self._write_line(meta, record)

        if self.write:
            self._write_after_iter()

        return gen2

    def __iter__(self):
        pass

    def _write_line(self, meta, line):
        # update count
        count = self.meta_count.get(meta, 0)
        self.meta_count[meta] = count + 1

        # write
        if self.write_mode == 'whole':
            meta = 'ALL'
        chunk = self.meta_dict.get(meta, [])

        if len(chunk) >= self.chunk_size:
            self._write_chunk(meta, chunk)
            self.meta_dict[meta] = [line]
        else:
            chunk.append(line)
            self.meta_dict[meta] = chunk

    def _write_chunk(self, meta, chunk):
        dynamic_fields = False
        if meta not in self.writer_dict:
            if meta == 'ALL':
                fns = [self.prefix, self.suffix]
            else:
                fns = [self.prefix, meta, self.suffix]
            fns = [x for x in fns if x]
            self.out_fn = os.path.join(self.out_dir, '.'.join(fns))
            if self.out_fn not in self.out_fns:
                self.out_fns.append(self.out_fn)
            writer = self.writer_cls(fn=self.out_fn)
            if self.write_headers:
                writer.write_headers()
            if self.write_fields:
                if not self.dynamic_fields:
                    writer.write_fields()
                else:
                    dynamic_fields = True
            self.writer_dict[meta] = writer
        else:
            writer = self.writer_dict[meta]

        writer.write_chunk(chunk, dynamic_fields=dynamic_fields)

    def _write_after_iter(self):
        self._write_last_chunk()
        if self.write_stats:
            self._write_stat()

    def _write_last_chunk(self):
        for meta, chunk in self.meta_dict.items():
            self._write_chunk(meta, chunk)

    def _write_stat(self):
        stat_fn = os.path.join(
            self.out_dir, self.prefix + '.' + self.suffix + '.stat')
        with open(stat_fn, 'w') as f:
            f.write('\t'.join(map(str, self.meta_count.keys())) + '\n')
            f.write('\t'.join(map(str, self.meta_count.values())) + '\n')



import os
from io import BytesIO
import matplotlib
import matplotlib.pyplot as plt

import svgwrite
from biotool.bplot import svgobject
from biotool.bplot import svgbase

plt.rc('mathtext', fontset='cm')
matplotlib.use('agg')


def measure_text(text, font_size=10):
    '''
    measure text length for Arial,
    ref: https://bl.ocks.org/tophtucker/62f93a4658387bb61e4510c37e2e97cf
    '''
    widths = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0.2796875,0.2765625,0.3546875,0.5546875,0.5546875,0.8890625,0.665625,0.190625,0.3328125,0.3328125,0.3890625,0.5828125,0.2765625,0.3328125,0.2765625,0.2765625,0.5546875,0.5546875,0.5546875,0.5546875,0.5546875,0.5546875,0.5546875,0.5546875,0.5546875,0.5546875,0.2765625,0.2765625,0.5828125,0.5828125,0.5828125,0.5546875,1.0140625,0.66875,0.665625,0.721875,0.721875,0.665625,0.609375,0.7765625,0.721875,0.2765625,0.5,0.665625,0.5546875,0.8328125,0.721875,0.7765625,0.665625,0.7765625,0.721875,0.665625,0.609375,0.721875,0.665625,0.94375,0.665625,0.665625,0.609375,0.2765625,0.2765625,0.2765625,0.46875,0.58125,0.3328125,0.5546875,0.5546875,0.5,0.5546875,0.5546875,0.3125,0.5546875,0.5546875,0.221875,0.2671875,0.5,0.221875,0.8328125,0.5546875,0.5546875,0.5546875,0.5546875,0.3453125,0.5,0.2765625,0.5546875,0.5,0.721875,0.5,0.5,0.5,0.3328125,0.259375,0.3328125,0.5828125]
    avg = 0.5273190789473683
    temps = [widths[ord(c)] if ord(c) < len(widths) else avg for c in text]
    l = sum(temps) * font_size
    return l


def tex2svg(latex, fontsize=100, dpi=300, output=BytesIO()):
    fig = plt.figure(figsize=(0.01, 0.01))
    fig.text(0, 0, latex, fontsize=fontsize)
    fig.savefig(output, dpi=dpi, transparent=True, format='svg',
                bbox_inches='tight', pad_inches=0.01, frameon=False)
    plt.close(fig)
    if isinstance(output, BytesIO):
        output.seek(0)
        return output.read()
    else:
        return


class LatexText(svgbase.SVGObject):

    def __init__(self, text, name, insert=(0, 0), size=(5, 5),
                 font_size=10,
                 anchor_object=None, anchor_pos='',
                 target_object=None, target_pos='',
                 **kwargs):

        self.x, self.y = insert
        self.width, self.height = size

        super(LatexText, self).__init__(insert, size,
                                        font_size=font_size,
                                        anchor_object=anchor_object,
                                        anchor_pos=anchor_pos,
                                        target_object=target_object,
                                        target_pos=target_pos,
                                        **kwargs)

        if not os.path.exists('latex'):
            os.makedirs('latex')
        fn = 'latex/{}.svg'.format(name)
        tex2svg(text, fontsize=font_size, dpi=300,
                output=fn)

        self.svgobj = svgwrite.image.Image(fn, self.insert, self.size,
                                           **kwargs)


class TextArea(svgobject.Container):

    def __init__(self, text, insert,
                 font_size=10,
                 **kwargs):
        self.texts = text.split('\n')
        self.font_size = font_size

        size = self._get_size()
        super(TextArea, self).__init__(insert, size, **kwargs)
        self.split_items = sum(self.split(split_pattern=(1, len(self.texts))), [])

    def render(self):
        for text, text_container in zip(self.texts, self.split_items):
            text_container = text_container.render()
            insert = 0, 0
            text_object = svgobject.Text(text, insert,
                                         parent=text_container, related_insert='center:middle').render()

            text_container.add(text_object)
            self._svg.add(text_container)

        return self._svg

    def _get_size(self):
        width = max(len(x) for x in self.texts)*8
        height = len(self.texts)*self.font_size
        size = width, height
        return size



# -*- coding: utf-8 -*-
"""
    simulate.fasta
    ~~~~~~~~~~~~~~

    

    @Copyright: (c) 2017-08 by Lingxi Chen (chanlingxi@gmail.com). 
    @License: LICENSE_NAME, see LICENSE for more details.
"""
from Bio import SeqIO


def get_fasta_total_length(fasta_fn):
    length = 0 
    for line in SeqIO.parse(open(fasta_fn, 'rU'), 'fasta'):
        length += len(line.seq)
    return length

   
    



# -*- coding: utf-8 -*-
"""
    simulate.math
    ~~~~~~~~~~~~~

    self-defined math function

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com). 
    @License: LICENSE_NAME, see LICENSE for more details.
"""
from numpy import random

def randint(low, high):
    if low == high: return low
    else:           return random.randint(low, high)



# -*- coding: utf-8 -*-

"""
    utils.complex_haplotype
    ~~~~~~~~~~~~~~~

    Local complex haplotype definition.

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
from genomic_region import GenomicRegion

class Interval(object):

    def __init__(self, pos1, pos2, index, notation='H'):
        self.start = min(pos1, pos2)
        self.end = max(pos1, pos2)
        self.index = index # start from 0
        self.notation = notation
        self.name = self.notation + str(self.index+1)

    def __cmp__(self, other):
        return self.index > other.index

    def __eq__(self, other):
        return self.start == other.start and self.end == other.end and self.name == other.na0
    def __repr__(self):
        return '%s[%s, %s]' % (self.name, str(self.start), str(self.end))
    def __str__(self):
        return str(self.name)
    def to_stat(self):
        return '[SEG]\t{}\t{}\t{}'.format(self.name, self.start, self.end)



class SVLink(object):

    def __init__(self, sv_name=None, up_pos=None, down_pos=None):
        self.sv_name = sv_name # 'DEL', 'INV', 'TD'
        self.up_pos = up_pos
        self.down_pos = down_pos

    def __eq__(self, other):
        return self.sv_name == other.sv_name and self.up_pos == other.up_pos and self.down_pos == other.down_pos

    def __repr__(self):
        if not self.sv_name and not self.up_pos and not self.down_pos: return ' %2s%3s%2s ' % ('','','')
        return ' %2s%3s%2s ' % (str(self.up_pos), self.sv_name, str(self.down_pos))
    def __str__(self):
        return ' %3s ' % (str(self.sv_name) if self.sv_name else '')

    @staticmethod
    def render_sv_link(new_region, last_region, sv_name=None):
        up_pos, down_pos = last_region.get_3prime_pos(), new_region.get_5prime_pos()
        if sv_name: return SVLink(sv_name, up_pos, down_pos)
        # infer sv name
        if new_region.strand != last_region.strand:                             sv_name = 'INV'
        elif new_region.strand == '+':
            if last_region.get_3prime_pos() < new_region.get_5prime_pos()-1:    sv_name = 'DEL'
            if last_region.get_3prime_pos() == new_region.get_5prime_pos()-1:   sv_name = '   '
            if last_region.get_3prime_pos() > new_region.get_5prime_pos()-1:    sv_name = 'TD'
        elif new_region.strand == '-':
            if last_region.get_3prime_pos() > new_region.get_5prime_pos()-1:    sv_name = 'DEL'
            if last_region.get_3prime_pos() == new_region.get_5prime_pos()-1:   sv_name = '   '
            if last_region.get_3prime_pos() < new_region.get_5prime_pos()-1:    sv_name = 'TD'
        return SVLink(sv_name, up_pos, down_pos)


class Contig(object):

    def __init__(self, contig=None):
        self.contig = contig or []

    def append_connection(self, region, sv_link=SVLink()):
        self.contig.append((sv_link, region))

    def __add__(self, other):
        contig = self.contig + other.contig
        return Contig(contig)
    def __len__(self):
        return self.contig.__len__()

    def get_contig_length(self):    return len(self.contig)
    def get_first_region(self):     return self.contig[0][1]
    def get_last_region(self):      return self.contig[-1][1]
    def get_first_sv_link(self):    return self.contig[0][0]
    def get_last_sv_link(self):     return self.contig[-1][0]

    def __getitem__(self, key): # index or slice object
        item = self.contig.__getitem__(key)
        if type(item) == tuple: item = [item]
        return Contig(item)

    def __delitem__(self, key): # index or slice object
        contig = self.__getitem__(key)
        self.contig.__delitem__(key)

    def pop(self, key): # index or slice object
        contig = self.__getitem__(key)
        self.__delitem__(key)
        return contig

    def __repr__(self):
        return ''.join([repr(sv_link) + repr(region) for sv_link, region in self.contig])

    def __str__(self):
        return ''.join([str(sv_link) + str(region) for sv_link, region in self.contig])

    def to_stat(self):
        return ','.join(region.full_name for _, region in self.contig)

    ####
    # transform contig to fasta #
    ####
    def get_fasta(self, ref_fasta, out_fasta=False):
        contig_header, contig_seq = '', ''
        for sv_link, region in self.contig:
            header, seq = region.get_fasta(ref_fasta, out_fasta)
            contig_header += header
            contig_seq += seq
        return contig_header, contig_seq

    @property
    def bp_length(self):
        return sum(region.bp_length for svlink, region in self.contig)

class TandemDuplicationContig(object):

    def __init__(self, repeat_contig, repeat_time=5, generate_type='forward', base_contig=Contig(), margin=0):
        self.repeat_contig = repeat_contig
        self.repeat_time = repeat_time
        self.generate_type = generate_type # 'backword' or 'forward'
        self.base_contig = base_contig # Empty if 'forward'
        self.margin = margin # graphical representation margin, 0 if 'forward'


    def get_last_region(self):
        return self.repeat_contig.get_last_region()

    def __repr__(self):
        base_str = repr(self.base_contig) + '\n' if self.generate_type == 'backward' else ''
        repeat_str = (repr(GenomicRegion())+repr(SVLink()))*self.margin + repr(self.repeat_contig)
        repeat_str_list = [repeat_str]*self.repeat_time
        return base_str + '\n'.join(repeat_str_list)

    def __str__(self):
        base_str = str(self.base_contig) + '\n' if self.generate_type == 'backward' else ''
        repeat_str = (' '*(9+1+3))*self.margin + ' '*self.margin + str(self.repeat_contig)
        repeat_str_list = [repeat_str]*self.repeat_time
        return base_str + '\n'.join(repeat_str_list)

    def to_stat(self):
        if self.generate_type == 'backward':
            res = '{},'.format(self.base_contig.to_stat())
        else:
            res = ''
        res += '({}):{}'.format(self.repeat_contig.to_stat(), self.repeat_time)
        return res

    ####
    # transform tandemduplication contig to fasta #
    ####
    def get_fasta(self, ref_fasta, out_fasta=False):
        base_contig_header, base_contig_seq = self.base_contig.get_fasta(ref_fasta)
        repeat_contig_header, repeat_contig_seq = self.repeat_contig.get_fasta(ref_fasta)
        tdc_header = base_contig_header
        tdc_seq = base_contig_seq + ''.join([repeat_contig_seq]*self.repeat_time)

        return tdc_header, tdc_seq

    @property
    def bp_length(self):
        bp_length = self.repeat_contig.bp_length*self.repeat_time
        if self.generate_type == 'backward':
            bp_length += self.base_contig.bp_length
        return bp_length

class ComplexHaplotype(object):

    def __init__(self, region, interval_list, is_local=True, is_inter=True,
            elongated_length=500000,
            ):
        self.region = region
        self.interval_list = interval_list

        self.contig_list = [Contig()]
        self.is_local = is_local
        self.is_inter = is_inter
        self._in_dup = False
        self.elongated_length = elongated_length

        self.up_elongated_region = self.region.slice(self.region.start-1, self.region.start-elongated_length, '+', 'UP','UP')
        self.down_elongated_region = self.region.slice(self.region.end+1, self.region.end+elongated_length, '+', 'DOWN','DOWN')
        self.elongated_region = self.region.slice(self.region.start-elongated_length, self.region.end+elongated_length, '+', 'elongated', 'elongated')

    ####
    # append methods #
    ####
    def append_contig(self, contig):
        self.contig_list.append(contig)

    def append_connection(self, region, sv_link=SVLink()):
        if type(self.get_last_contig()) != Contig:
            self.contig_list.append(Contig())
        self.get_last_contig().append_connection(region, sv_link)
    ####
    # get methods #
    ####
    def get_last_contig(self): return self.contig_list[-1]
    def get_last_region(self): return self.get_last_contig().get_last_region()
    def pop_subcontig_from_last_contig(self, key):
        if type(self.get_last_contig()) != Contig: return Contig()
        subcontig = self.get_last_contig().pop(key)
        return subcontig

    def __eq__(self, other):
        return self.haplotype == other.haplotype
    def __repr__(self):
        res = ''
        for contig1, contig2 in zip(['']+self.contig_list, self.contig_list+[Contig()]):
            delimiter = '' if type(contig1) == Contig and type(contig2) == TandemDuplicationContig and contig2.generate_type == 'backward' else '\n'
            res += delimiter + repr(contig2)
        return res
    def __str__(self):
        res = self.up_elongated_region.full_name
        for contig1, contig2 in zip(['']+self.contig_list, self.contig_list+[Contig()]):
            delimiter = '' if type(contig1) == Contig and type(contig2) == TandemDuplicationContig and contig2.generate_type == 'backward' else '\n'
            res += delimiter + str(contig2)
        res += self.down_elongated_region.full_name
        return res
    def to_stat(self):
        res = []
        # segment_list
        res.append(self.up_elongated_region.to_stat())
        res.append(self.interval_list.to_stat())
        res.append(self.down_elongated_region.to_stat())
        # length
        res.append('[LENGTH]\t{}\n'.format(self.bp_length))
        # hap
        res.append('[COMPLEX_HAPLOTYPE]\n'+
                self.up_elongated_region.full_name + ',' +
                ','.join(contig.to_stat() for contig in self.contig_list) +
                self.down_elongated_region.full_name
                )
        res.append('[GRAPH]\n'+str(self))
        return '\n'.join(res)

    @property
    def bp_length(self):
        return sum(contig.bp_length for contig in self.contig_list) + self.elongated_length*2


    ####
    # write haplotype to fasta #
    ####
    def write_to_fasta(self, ref_fasta, out_fasta, mod='wb'):

        hap_seq = ''

        header, seq = self.up_elongated_region.get_fasta(ref_fasta)
        hap_seq += seq

        for contig in self.contig_list:
            header, seq = contig.get_fasta(ref_fasta)
            hap_seq += seq
        hap_header = self.elongated_region.to_fasta_header()

        header, seq = self.up_elongated_region.get_fasta(ref_fasta)
        hap_seq += seq


        with open(out_fasta, mod) as out:   out.write(hap_header + '\n' + hap_seq + '\n')
        return hap_header, hap_seq

    def write_to_stat(self, stat_fn, mod='wb'):
        with open(stat_fn, mod) as out:   out.write(self.to_stat()+'\n')

if __name__ == "__main__":
    h = ComplexHaplotype(GenomicRegion('chr1','0','10000000','+','test'))
    '''
    h.append_connection(GenomicRegion('chr1','a','b','+'))
    h.append_connection(GenomicRegion('chr1','c','d','+'), SVLink('DEL','b','c'))

    base_c = Contig()
    base_c.append_connection(GenomicRegion('chr1','e','f','+'), SVLink('DEL','d','e'))
    base_c.append_connection(GenomicRegion('chr1','d','f','-'), SVLink('INV','f','f'))
    base_c.append_connection(GenomicRegion('chr1','b','c','-'), SVLink('DEL','d','c'))
    base_c.append_connection(GenomicRegion('chr1','g','h','+'), SVLink('INV','b','g'))
    repeat_c = Contig()
    repeat_c.append_connection(GenomicRegion('chr1','e','f','+'), SVLink('TD','h','e'))
    repeat_c.append_connection(GenomicRegion('chr1','d','f','-'), SVLink('INV','f','f'))
    repeat_c.append_connection(GenomicRegion('chr1','b','c','-'), SVLink('DEL','d','c'))
    repeat_c.append_connection(GenomicRegion('chr1','g','h','+'), SVLink('INV','b','g'))
    tdc = TandemDuplicationContig(repeat_c,generate_type='backward',base_contig=base_c, margin=2)
    h.append_contig(tdc)

    repeat_c = Contig()
    repeat_c.append_connection(GenomicRegion('chr1','i','j','+'), SVLink('DEL','h','i'))
    repeat_c.append_connection(GenomicRegion('chr1','b','h','+'), SVLink('TD','j','b'))
    tdc=TandemDuplicationContig(repeat_c, repeat_time=3)
    h.append_contig(tdc)
    h.append_connection(GenomicRegion('chr1','h0','k','+'), SVLink('','h','h'))
    h.append_connection(GenomicRegion('chr1','h1','k','+'), SVLink('','h','h'))
    h.append_connection(GenomicRegion('chr1','h2','k','+'), SVLink('','h','h'))
    h.append_connection(GenomicRegion('chr1','h3','k','+'), SVLink('','h','h'))
    '''
    # test for get_fasta
    h.append_connection(GenomicRegion('chr1','100010','100020','+'))
    base_c = Contig()
    base_c.append_connection(GenomicRegion('chr1','100030','100040','+'), SVLink('DEL','100020','100030'))
    base_c.append_connection(GenomicRegion('chr1','100025','100035','+'), SVLink('TD','100040','100025'))
    base_c.append_connection(GenomicRegion('chr1','100035','100025','-'), SVLink('INVS','100035','100035'))
    tdc = TandemDuplicationContig(base_c, repeat_time=2)
    h.append_contig(tdc)


    repeat_c = Contig()
    repeat_c.append_connection(GenomicRegion('chr1','100030','100040','+'), SVLink('INV','100025','100030'))
    repeat_c.append_connection(GenomicRegion('chr1','100025','100035','+'), SVLink('TD','100040','100025'))
    repeat_c.append_connection(GenomicRegion('chr1','100035','100025','-'), SVLink('INVS','100035','100035'))
    tdc = TandemDuplicationContig(repeat_c, generate_type='backward', base_contig=base_c, margin=0, repeat_time=3)
    h.append_contig(tdc)
    #print h
    #print '-'*20

    ref_fasta = '/home/BIOINFO_DATABASE/reference/genome_DNA/Homo_sapiens/hg19/BWA_GATK_index/hg19.fa'
    out_fasta = 'test'
    header, seq = h.get_fasta(ref_fasta, out_fasta)
    #print header
    #print seq



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    clustering
    ~~~~~~~~~~~~~~~~~~~~



    @Copyright: (c) 2018-09 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""

from sklearn import decomposition
import seaborn as sns
import matplotlib.pyplot as plt
sns.set(color_codes=True)
plt.switch_backend('agg')


def pca(X, y=None, n_components=2):
    pca = decomposition.PCA(n_components=n_components)
    pca.fit(X)
    X = pca.transform(X)
    return X


def draw_hierarchical_clustering(out_prefix, X, y=None,
                                 method='average', metric='euclidean',
                                 standard_scale=None, z_score=None):
    if y is not None:
        lut = dict(zip(y.unique(), "rb"))
        row_colors = y.map(lut)
        row_colors.index = X.index
        sns.clustermap(X, row_colors=row_colors,
                       method=method, metric=metric,
                       standard_scale=standard_scale, z_score=z_score,
                       )
    else:
        sns.clustermap(X,
                       method=method, metric=metric,
                       standard_scale=standard_scale, z_score=z_score,
                       )

    plt.suptitle('Hierarchical Clustering:' + method + '.' + metric)
    plt.savefig(out_prefix + '_' + method + '_' + metric + '.png', format='png')


def draw_PC2(out_prefix, X, y):
    plt.figure(figsize=(8, 6))
    tumor = X[0:8]
    normal = X[8:16]
    plt.plot(tumor['PC1'], tumor['PC2'], '*r', label='Tumor')
    plt.plot(normal['PC1'], normal['PC2'], '*b', label='Normal')

    plt.legend(loc='upper right')
    plt.xlabel('PC1')
    plt.ylabel('PC2')
    plt.suptitle('PC2 Cluster')
    plt.savefig(out_prefix + '_PC2.png', format='png')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    bash
    ~~~~~~~~~~~~~~

    bash tools to access directory

    @Copyright: (c) 2018-07 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""

import os


def get_files(in_dir, fn_suffix=''):

    in_fns = []
    for root, _, fns in os.walk(in_dir):
        if not fns:
            continue
        for fn in fns:
            if not fn.endswith(fn_suffix):
                continue
            in_fns.append(os.path.join(root, fn))

    return in_fns



# -*- coding: utf-8 -*-
"""
    utils.genomic_region
    ~~~~~~~~~~~~

    Genomic region definition.

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
import copy
import pybedtools
import os, sys
import sequence

from bioutensil import constants

class GenomicRegion(object): # [start, end]; length = end - start + 1

    def __init__(self, chrom=None, pos1=None, pos2=None, strand=None, name=None, code_name=None, full_name=None):
        if pos1 and pos1 == pos2: raise Exception('In class GenomicRegion, pos1 %s cannot equal to pos2 %s' % (str(pos1), str(pos2)))
        self.chrom = chrom
        try: # [start, end] closed interval
            self.start = min(int(pos1), int(pos2))
            self.end = max(int(pos1), int(pos2))
        except Exception as e:
            self.start = min(pos1, pos2)
            self.end = max(pos1, pos2)
        self.strand = strand
        self.name = name
        self.code_name = code_name
        self.full_name = full_name


    def slice(self, pos1, pos2, new_strand, code_name, full_name):
        new_region = copy.copy(self)
        new_region.start = min(pos1, pos2)
        new_region.end = max(pos1, pos2)
        new_region.strand = new_strand
        new_region.code_name = code_name
        new_region.full_name = full_name
        return new_region

    def get_3prime_pos(self):
        if self.strand == '+':      return self.end
        if self.strand == '-':      return self.start
    def get_5prime_pos(self):
        if self.strand == '+':      return self.start
        if self.strand == '-':      return self.end

    @property
    def bp_length(self):
        return self.end - self.start + 1

    def __eq__(self, other):
        return self.chrom == other.chrom and self.start == other.start and self.end == other.end and self.name == other.name and self.strand == other.strand and self.code_name == other.code_name and self.full_name == other.full_name

    ####
    # representation #
    ####
    def to_fasta_header(self):
        return '>%s::%s::%s:%s-%s' % (self.name, self.strand, self.chrom, str(self.start), str(int(self.end)+1))

    def to_bed(self):
        if self.strand == '+':  start, end = self.start, int(self.end) + 1
        else:                   start, end = int(self.start) - 1, self.end
        return '\t'.join(str(v) if v else '' for v in [self.chrom, start, end, self.name])
    def __repr__(self):
        if not self.start: return '%s %s %s ' % (' ','  ','  ')
        code_name = str(self.code_name) if self.code_name else ''
        if self.strand == '+':
            return '%s[%2s,%2s]%s' % (str(self.strand), str(self.start), str(self.end), code_name)
        if self.strand == '-':
            return '%s[%2s,%2s]%s' % (str(self.strand), str(self.end), str(self.start), code_name)
    def __str__(self):
        return '%6s' % (str(self.code_name) if self.code_name else '')

    def to_stat(self):
        return '[SEG]\t{}\t{}\t{}'.format(self.code_name, self.start, self.end)
    ####
    # get region fasta sequence #
    ####
    def get_fasta(self, ref_fasta, out_fasta=False):
        tmp = 'tmp_fasta'
        if out_fasta: tmp=out_fasta
        bedtool = pybedtools.BedTool(self.to_bed(),from_string=True)
        bedtool.sequence(fi=ref_fasta, fo=tmp, name=True)

        header, seq = open(bedtool.seqfn, 'rb').read().split()
        header, seq = header.strip(), seq.strip()
        if self.strand == '-':
            seq = sequence.reverse_complement(seq)
        if out_fasta:   # return file
            with open(out_fasta, 'wb') as out:
                out.write(header + '\n')
                out.write(seq + '\n')
            return
        else:           # return string
            os.remove(tmp)
            return header, seq

    @staticmethod
    def safe_genomic_region(chrom, start, end):
        if start < 0: start = 0   # start from 0???
        max_end = constants.hg19_fai_bp[chrom]    # hg19 !!!????
        if end > max_end: end = max_end
        return chrom, start, end

if __name__ == "__main__":
    ref_fasta = '/home/BIOINFO_DATABASE/reference/genome_DNA/Homo_sapiens/hg19/BWA_GATK_index/hg19.fa'
    ####
    # region1 and region2 are the same!!! #
    ####
    region1 = GenomicRegion('chr1','100010','100020','+')
    region2 = GenomicRegion('chr1','100010','100020','-')
    '''
    print region1
    header1, seq1 = region1.get_fasta(ref_fasta)
    print seq1, sequence.reverse_complement(seq1)
    print region2
    header2, seq2 = region2.get_fasta(ref_fasta)
    print seq2, sequence.reverse_complement(seq2)
    '''




# -*- coding: utf-8 -*-
"""
    psl_tran
    ~~~~~~~~~~~~~~

    PSL transcript class

    @Copyright: (c) 2018-04 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
import pysam

from bioutensil import genomic_region
from bioutensil import gene


class Reader(object):

    def __init__(self, fn, gene_list=None):
        self._fn = fn
        self.gene_list = gene.read_gene_list_from_dir(gene_list)

    def fetch(self, chrom, pos1, pos2):
        if pos1 > pos2:
            start, end = pos2, pos1
        else:
            start, end = pos1, pos2
        tbx = pysam.TabixFile(self._fn)
        try:
            for i, row in enumerate(tbx.fetch(chrom, start, end)):
                yield Transcript(row)
        except Exception:
            return

    def fetch_genes_trans(self, chrom, pos1, pos2):
        groups = []
        gene = None
        for i, tran in enumerate(self.fetch(chrom, pos1, pos2)):
            if gene != tran.gene:
                if groups:
                    yield groups
                groups = []
            groups.append(tran)
            gene = tran.gene
        # yield last group
        if groups:
            yield groups

    def fetch_genes_tran_by_point(self, chrom, point):
        return self.fetch_genes_tran_by_region(chrom, point, point+1)

    def fetch_genes_tran_by_region(self, chrom, pos1, pos2):
        genes_trans = self.fetch_genes_trans(chrom, pos1, pos2)
        for i, gene_trans in enumerate(genes_trans):
            tran = self.choose_tran_from_group(gene_trans)
            if not tran:
                continue
            for group, genes in self.gene_list.items():
                if tran.gene in genes:
                    tran.group.append(group)
            yield tran

    def fetch_genes_tran_by_seg(self, chrom, pos, strand, seg_extend=500000):
        if strand == '+':
            start, end = pos - seg_extend, pos
        else:
            start, end = pos, pos + seg_extend
        chrom, start, end = genomic_region.GenomicRegion.safe_genomic_region(chrom, start, end)
        return self.fetch_genes_tran_by_region(chrom, start, end)

    @staticmethod
    def choose_tran_from_group(trans,
                               default_biotypes=['protein_coding', 'lincRNA'],  # ,'miRNA'],
                               other_biotypes=[]):
        # choose one transcript to represent the gene transcript
        accepted_biotypes = default_biotypes + other_biotypes

        protein_001, protein_201, protein_min = None, None, None
        other_001, other_201, other_min = None, None, None
        protein_name_id_min, other_name_id_min = '999', '999'
        for t in trans:
            if t.biotype not in accepted_biotypes:
                continue
            name_id = t.name.split('-')[-1]
            if 'protein_encoding' in t.biotype:
                if name_id == '001':
                    protein_001 = t
                if name_id == '201':
                    protein_201 = t
                if name_id < protein_name_id_min:
                    protein_min = t
                    protein_name_id_min = name_id
            else:
                if name_id == '001':
                    other_001 = t
                if name_id == '201':
                    other_201 = t
                if name_id < other_name_id_min:
                    other_min = t
                    other_name_id_min = name_id

        if protein_001:
            return protein_001
        if protein_201:
            return protein_201
        if protein_min:
            return protein_min
        if other_001:
            return other_001
        if other_201:
            return other_201
        if other_min:
            return other_min
        return None


class Transcript(object):

    def __init__(self, line):

        self._fields = 'source,version,col3,col4,start_condon,end_condon,protein_ens_id,col8,strand,name,ens_id,gtf_name,biotype,chrom,cytoband,left,right,exon_num,exon_lens,cds_info,exon_starts,gene'.split(',')

        self._parse(line)

    def _set(self, source=None, version=None, col3=None, col4=None,
             start_condon=None, end_condon=None, protein_ens_id=None, col8=None,
             strand=None, name=None, ens_id=None, gtf_name=None,
             biotype=None, chrom=None, cytoband=None, left=None,
             right=None, exon_num=None, exon_lens=None, cds_info=None,
             exon_starts=None, gene=None,
             extend=20000,
             group=None):
        self.source = source
        self.version = version
        self.start_condon = tuple(int(x) for x in start_condon.split(',') if x)
        self.end_condon = tuple(int(x) for x in end_condon.split(',') if x)
        self.protein_ens_id = protein_ens_id
        self.col8 = col8
        self.strand = strand
        self.name = name
        self.ens_id = ens_id
        self.gtf_name = gtf_name
        self.biotype = biotype
        self.chrom = chrom
        self.cytoband = cytoband
        self.left = int(left) + 1   # first exon start
        self.right = int(right)     # last exon end
        self.exon_num = exon_num
        self.exon_lens = list(int(x) for x in exon_lens.split(',') if x)
        self.cds_info = cds_info
        self.exon_starts = list(int(x) for x in exon_starts.split(',') if x)
        self.gene = gene

        self.extend = int(extend)
        self.left_extend = self.left - self.extend
        self.right_extend = self.right + self.extend

        self.group = group or []

    def _parse(self, line):  # parse line to a record object
        args = dict(zip(self._fields, line.strip().split()))
        self._set(**args)

    def __str__(self):
        return '{} {} {} {} {} {} {}'.format(self.chrom, self.left, self.right, self.name, self.biotype, self.gene, self.strand)

    @staticmethod
    def _in_region(point, region):  # left and right are all in region
        if not region:
            return False
        left, right = region
        return left <= point and right >= point

    def in_exons(self, point):
        for i, exon in enumerate(self.exons):
            if self._in_region(point, exon):
                return i+1
        return 0

    def in_introns(self, point):
        for i, intron in enumerate(self.introns):
            if self._in_region(point, intron):
                return i+1
        return 0

    def in_utr5(self, point):
        return self._in_region(point, self.utr5)

    def in_utr3(self, point):
        return self._in_region(point, self.utr3)

    def in_promoter(self, point):
        return self._in_region(point, self.promoter)

    def in_terminator(self, point):
        return self._in_region(point, self.terminator)

    def in_upstream(self, point):  # 5' upstream
        if self.strand == '+':
            return point < self.left_extend
        if self.strand == '-':
            return point > self.right_extend

    def in_downstream(self, point):  # 3' downstream
        if self.strand == '+':
            return point > self.right_extend
        if self.strand == '-':
            return point < self.left_extend

    def in_genetic(self, point):
        return self.in_exons(point) \
                or self.in_introns(point) \
                or self.in_utr5(point) \
                or self.in_utr3(point)

    def mark_functional_region(self, point):
        g = 'GENETIC'
        ig = 'INTERGENETIC'
        reg = 'REGULATION'

        if self.in_utr5(point):
            return g, "utr5"
        if self.in_utr3(point):
            return g, "utr3"
        # utr5 and utr3 are inside exon
        exon = self.in_exons(point)
        if exon:
            return g, 'e{}'.format(exon)
        intron = self.in_introns(point)
        if intron:
            return g, 'i{}'.format(intron)
        if self.in_promoter(point):
            return reg, "promoter"
        if self.in_terminator(point):
            return reg, "terminator"
        if self.in_upstream(point):
            return ig, "upstream"
        if self.in_downstream(point):
            return ig, "downstream"
        return None, None

    def mark_exon_distance(self, point):
        if self.in_exons(point):
            return 0
        intron_i = self.in_introns(point)
        if intron_i:
            l, r = self.get_intron(intron_i)
            l, r = l-1, r+1
            return min(point-l, r-point)
            ''' in intergenetic or regulation area '''
        else:
            return min(abs(point-self.left), abs(point-self.right))
        return None

    def get_exon(self, i):
        exons = self.exons
        if abs(i) > len(exons) or abs(i) < 1:
            return None, None
        if i > 0:
            return exons[i-1]
        if i < 0:
            return exons[i]

    def get_intron(self, i):
        introns = self.introns
        if abs(i) > len(introns) or abs(i) < 1:
            return None, None
        if i > 0:
            return introns[i-1]
        if i < 0:
            return introns[i]

    @property
    def exons(self):  # start = start + 1, end = start + len, len is the number of base pairs
        exon_starts = [start + 1 for start in self.exon_starts]
        exon_ends = [start + l for start, l in zip(self.exon_starts, self.exon_lens)]
        if self.strand == '-':
            exon_starts = exon_starts[::-1]
            exon_ends = exon_ends[::-1]
        return list(zip(exon_starts, exon_ends))

    @property
    def introns(self):
        if self.strand == '+':
            return [(left[-1]+1, right[0]-1) for left, right in zip(self.exons[:-1], self.exons[1:])]
        if self.strand == '-':
            return [(left[-1]+1, right[0]-1) for right, left in zip(self.exons[:-1], self.exons[1:])]

    @property
    def utr5(self):
        if not self.start_condon:
            return tuple()
        if self.strand == '+':
            return (self.left, self.start_condon[0] - 1)
        if self.strand == '-':
            return (self.start_condon[-1] + 1, self.right)

    @property
    def utr3(self):
        if not self.end_condon:
            return tuple()
        if self.strand == '+':
            return (self.end_condon[-1] + 1, self.right)
        if self.strand == '-':
            return (self.left, self.end_condon[0] - 1)

    @property
    def promoter(self):
        if self.strand == '+':
            return (self.left_extend, self.left - 1)
        if self.strand == '-':
            return (self.right + 1, self.right_extend)

    @property
    def terminator(self):
        if self.strand == '-':
            return (self.left_extend, self.left - 1)
        if self.strand == '+':
            return (self.right + 1, self.right_extend)


'''
    def test(self, start=100, end=200):
        print 'left', self.left
        print 'right', self.right
        print self.right - self.left
        print '--- exons'
        print 'exon_num', self.exon_num
        print 'exon_starts', self.exon_starts
        print 'exon_lens', self.exon_lens
        print 'exons', self.exons
        print '--- intron'
        print self.introns

        print '--- utr'
        print 'start condon', self.start_condon
        print 'end condon', self.end_condon
        print 'cds_info', self.cds_info   #????
        print '3utr', self.utr3
        print '5utr', self.utr5
        print 'promoter', self.promoter
        print 'terminator', self.terminator
        if self.start_condon and self.end_condon:
            print 'start condon - end condon', self.start_condon[0] - self.end_condon[-1]
        print '-----distance----'
        point = self.left - 1000
        print 'point', point, 'distance', self.mark_exon_distance(point)
        point = self.right + 1000
        print 'point', point, 'distance', self.mark_exon_distance(point)
        point = 1167272
        print 'point', point, 'distance', self.mark_exon_distance(point)
        point = 1159349 + 99
        print 'point', point, 'distance', self.mark_exon_distance(point)
        point = 1163847 - 99
        print 'point', point, 'distance', self.mark_exon_distance(point)

if __name__ == "__main__":
    line = 'ensembl NA      0       0       1164171,1164172,1164173,        1156079,1156080,1156081,        ENSP00000444451 0       -       SDF4-201        ENST00000545427 SDF4-201        protein_coding  chr1    p36.33  1156078 1167370 5       32,114,137,479,99,   1156081(29),1158623(114),1159211(137),1163847(326),     1156078,1158623,1159211,1163847,1167271,        SDF4'
    print line
    tran = Transcript(line)
    tran.test()

'''



# -*- coding: utf-8 -*-
"""
    utils.myprint
    ~~~~~~~~~~~~~

    my pretty print functions

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com). 
    @License: LICENSE_NAME, see LICENSE for more details.
"""

#######################################################################
#                            Matrix Print                             #
#######################################################################
import numpy as np

def mprint(matrix):
    for row in matrix:
        for item in row:
            item = [ [x for x in line] for line in item.split('\n') ]
            combine = np.hstack((combine, np.array(item)))
            print(combine)
            print('----------------')



if __name__ == "__main__":
    m =[[
"""-----
| a b |
|  +  |
-----""",
"""DEL
---"""
            ],
    ]
    mprint(m)




import svgwrite


class SVGObject():
    def __init__(self, insert, size,
                 anchor_object=None, anchor_pos='',
                 target_object=None, target_pos='',
                 **kwargs
                 ):

        self.insert = insert
        self.size = size

        self.anchor_object = anchor_object
        self.anchor_pos = anchor_pos
        self.target_object = target_object
        self.target_pos = target_pos

        if self.anchor_object:
            self._update_insert()
        if self.target_object:
            self._update_size()

    def _update_insert(self):
        x, y = self.anchor_object.x, self.anchor_object.y
        width, height = self.anchor_object.width, self.anchor_object.height

        if 'iup' in self.anchor_pos:
            self.y = y
        if 'oup' in self.anchor_pos:
            self.y = y - self.height

        if 'center' in self.anchor_pos:
            self.y = y + height/2 - self.height/2

        if 'ibottom' in self.anchor_pos:
            self.y = y + height - self.height
        if 'obottom' in self.anchor_pos:
            self.y = y + height

        if 'ileft' in self.anchor_pos:
            self.x = x
        if 'oleft' in self.anchor_pos:
            self.x = x - self.width

        if 'middle' in self.anchor_pos:
            self.x = x + width/2 - self.width/2

        if 'iright' in self.anchor_pos:
            self.x = x + width - self.width
        if 'oright' in self.anchor_pos:
            self.x = x + width

        self.insert = (self.x, self.y)

    def _update_size(self):
        x, y = self.target_object.x, self.target_object.y
        width, height = self.target_object.width, self.target_object.height

        if 'iup' in self.target_pos:
            self.y2 = y
        if 'oup' in self.target_pos:
            self.y2 = y - self.height

        if 'center' in self.target_pos:
            self.y2 = y + height/2 - self.height/2

        if 'ibottom' in self.target_pos:
            self.y2 = y + height - self.height
        if 'obottom' in self.target_pos:
            self.y2 = y + height

        if 'ileft' in self.target_pos:
            self.x2 = x
        if 'oleft' in self.target_pos:
            self.x2 = x - self.width

        if 'middle' in self.target_pos:
            self.x2 = x + width/2 - self.width/2

        if 'iright' in self.target_pos:
            self.x2 = x + width - self.width
        if 'oright' in self.target_pos:
            self.x2 = x + width

        self.size = (self.x2 - self.x, self.y2 - self.y)


class Rect(SVGObject):
    def __init__(self, insert, size,
                 anchor_object=None, anchor_pos='',
                 target_object=None, target_pos='',
                 **kwargs):

        self.x, self.y = insert
        self.width, self.height = size

        super(Rect, self).__init__(insert, size,
                                   anchor_object=anchor_object,
                                   anchor_pos=anchor_pos,
                                   target_object=target_object,
                                   target_pos=target_pos,
                                   **kwargs)
        self.svgobj = svgwrite.shapes.Rect(self.insert, self.size, **kwargs)


class Circle(SVGObject):
    def __init__(self, insert, size,
                 anchor_object=None, anchor_pos='',
                 target_object=None, target_pos='',
                 **kwargs):

        self.x, self.y = insert
        self.width, self.height = size

        super(Circle, self).__init__(insert, size,
                                     anchor_object=anchor_object,
                                     anchor_pos=anchor_pos,
                                     target_object=target_object,
                                     target_pos=target_pos,
                                     **kwargs)

        self.r = size[0]/2.0
        center = self.insert[0] + self.r, self.insert[1] + self.r
        self.svgobj = svgwrite.shapes.Circle(center, self.r, **kwargs)


class Line(SVGObject):
    def __init__(self, insert, size,
                 anchor_object=None, anchor_pos='',
                 target_object=None, target_pos='',
                 **kwargs):
        self.x, self.y = insert
        self.width, self.height = size

        super(Line, self).__init__(insert, size,
                                   anchor_object=anchor_object,
                                   anchor_pos=anchor_pos,
                                   target_object=target_object,
                                   target_pos=target_pos,
                                   **kwargs)

        width, height = self.size
        end = self.x + width, self.y + height
        self.svgobj = svgwrite.shapes.Line(self.insert, end, **kwargs)


class Text(SVGObject):

    def __init__(self, text, insert=(0, 0),
                 font_size=10, font_family='Arial', font_style='normal',
                 anchor_object=None, anchor_pos='',
                 target_object=None, target_pos='',
                 **kwargs):

        self.x, self.y = insert
        self.width, self.height = 0, 0

        super(Text, self).__init__(insert, None,
                                   font_size=font_size,
                                   font_family=font_family,
                                   font_style=font_style,
                                   anchor_object=anchor_object,
                                   anchor_pos=anchor_pos,
                                   target_object=target_object,
                                   target_pos=target_pos,
                                   **kwargs)

        self.svgobj = svgwrite.text.Text(text, self.insert,
                                         font_size=font_size,
                                         font_family=font_family,
                                         font_style=font_style,
                                         dy=self.dy, style=self.style,
                                         **kwargs)

    def _update_insert(self):
        # font render in up-right direction
        SVGObject._update_insert(self)
        self.dy = ['1em']
        self.style = 'text-anchor: start'

        if 'iup' in self.anchor_pos:
            self.dy = ['1em']
        if 'oup' in self.anchor_pos:
            self.dy = ['-0.3em']

        if 'center' in self.anchor_pos:
            self.dy = ['0.3em']

        if 'ibottom' in self.anchor_pos:
            self.dy = ['-0.3em']
        if 'obottom' in self.anchor_pos:
            self.dy = ['1em']

        offset = 2
        if 'ileft' in self.anchor_pos:
            self.style = 'text-anchor: start'
            self.x += offset
        if 'oleft' in self.anchor_pos:
            self.style = 'text-anchor: end'
            self.x -= offset

        if 'iright' in self.anchor_pos:
            self.style = 'text-anchor: end'
            self.x -= offset
        if 'oright' in self.anchor_pos:
            self.style = 'text-anchor: start'
            self.x += offset

        if 'middle' in self.anchor_pos:
            self.style = 'text-anchor: middle'

        self.insert = self.x, self.y


class LinearGradientRect(Rect):

    def __init__(self, insert, size, start_color, start_opacity, end_color, end_opacity, **extra):
        super(LinearGradientRect, self).__init__(insert, size, **extra)

        self._start_color = start_color
        self._start_opacity = start_opacity
        self._end_color = end_color
        self._end_opacity = end_opacity

    def render(self):
        horizontal_gradient = self._dwg.linearGradient((0, 0), (1, 0))
        horizontal_gradient.add_stop_color(0, self._start_color, self._start_opacity)
        horizontal_gradient.add_stop_color(1, self._end_color, self._end_opacity)

        rect = self._dwg.rect(self.insert, self.size, fill=horizontal_gradient.get_paint_server(default='currentColor'))

        return horizontal_gradient, rect


class Container(SVGObject):
    def __init__(self, insert, size,
                 h_margin=0, v_margin=0,
                 caption=None,
                 caption_pos='',
                 fill='none',
                 border_stroke='none',
                 stroke='none',
                 anchor_object=None,
                 anchor_pos='',
                 target_object=None,
                 target_pos='',
                 **kwargs):

        self.h_margin = h_margin
        self.v_margin = v_margin
        width, height = size
        size = width+2*h_margin, height+2*v_margin

        self.caption = caption
        self.caption_pos = caption_pos
        self.x, self.y = size
        self.width, self.height = size

        super(Container, self).__init__(insert, size,
                                        anchor_object=anchor_object,
                                        anchor_pos=anchor_pos,
                                        target_object=target_object,
                                        target_pos=target_pos,
                                        **kwargs)
        self.x, self.y = self.size
        self.width, self.height = self.size

        self.svgobj = svgwrite.container.SVG(self.insert, self.size,
                                             stroke='none',
                                             **kwargs)  # update the data
        self.rect = Rect((0, 0), self.size,
                         fill=fill,
                         stroke=border_stroke,
                         **kwargs)  # update the data
        self.svgobj.add(self.rect.svgobj)

        if caption:
            self._draw_caption()

    def _draw_caption(self):
        label = Text(self.caption, (0, 0),
                     font_size=8, font_weight='bold',
                     anchor_object=self.rect,
                     anchor_pos=self.caption_pos)
        self.svgobj.add(label.svgobj)

    def split(self, split_pattern=None, split_ratio=None):
        width, height = self.size
        if split_pattern:
            split_x, split_y = split_pattern
            split_ratio_x = [1.0/split_x] * split_x
            split_ratio_y = [1.0/split_y] * split_y
        if split_ratio:
            split_ratio_x, split_ratio_y = split_ratio
            split_ratio_x = list(map(int, split_ratio_x.split(':')))
            split_ratio_y = list(map(int, split_ratio_y.split(':')))
            split_ratio_x = list(map(lambda n: 1.0*n/sum(split_ratio_x), split_ratio_x))
            split_ratio_y = list(map(lambda n: 1.0*n/sum(split_ratio_y), split_ratio_y))

        items = []
        child_y = 0
        for j in range(len(split_ratio_y)):
            child_height = height*split_ratio_y[j]

            line = []
            child_x = 0
            for i in range(len(split_ratio_x)):
                child_width = width*split_ratio_x[i]

                insert = child_x, child_y
                size = child_width, child_height
                child = Container(insert, size)
                line.append(child)

                child_x += child_width

            items.append(line)
            child_y += child_height
        return items



import svgwrite
from bplot import svgbase
import math


class SVGObject(object):

    def __init__(self,
            stroke = 'black', stroke_width = 1,
            stroke_dasharray=[], stroke_lines=['outline'], stroke_pattern=[1,1,1,1],
            fill='white', opacity=1,
            ):
        self._dwg = svgwrite.Drawing()

        self.stroke = stroke
        self.stroke_width = stroke_width
        self._stroke_lines = stroke_lines
        self._stroke_pattern = stroke_pattern
        if stroke_lines or stroke_pattern:
            self._draw_stroke_lines()

        self.fill = fill
        self.opacity = opacity


class Polygon(SVGObject):
    def __init__(self, points, **extra):
        self.points = points
        super(Polygon, self).__init__(**extra)


    def render(self):
        polygon = self._dwg.polygon(self.points,
                stroke=self.stroke, stroke_width=self.stroke_width, stroke_dasharray=self.stroke_dasharray,
                fill=self.fill, opacity=self.opacity)
        return polygon

    def _draw_stroke_lines(self):
        if 'outline' in self._stroke_lines: self._stroke_lines += ['up','right','bottom','left']
        pattern = [int(x in self._stroke_lines) for x in ['up','right','bottom','left']]
        if self._stroke_pattern: pattern = self._stroke_pattern

        distance = lambda x, y: math.sqrt(math.pow(x[0]-y[0], 2) + math.pow(x[1]-y[1], 2))
        size = [distance(x,y) for x, y in zip(self.points, self.points[1:]+[self.points[0]])]
        stroke_dasharray = []
        prev_pattern = -1
        element = 0
        for x, size in zip(pattern, size):
            if prev_pattern == x or prev_pattern == -1:
                element += size
            else: # new
                # update prev
                stroke_dasharray.append(element)
                element = size
            prev_pattern = x

        stroke_dasharray.append(element)

        if pattern[0] != 1:
            stroke_dasharray = [0] + stroke_dasharray

        self.stroke_dasharray=','.join(map(str,stroke_dasharray))


class Arrow(svgbase.Container):

    def __init__(self, insert, size,
                 border_stroke='none',
                 anchor_object=None,
                 anchor_pos='',
                 **kwargs):
        super(Arrow, self).__init__(insert, size,
                                    anchor_object=anchor_object,
                                    anchor_pos=anchor_pos,
                                    border_stroke=border_stroke,
                                    **kwargs)

        line = svgbase.Line((0, 0), (0, 0),
                            anchor_object=self.rect,
                            anchor_pos='middle:iup',
                            target_object=self.rect,
                            target_pos='middle:ibottom',
                            **kwargs)
        self.svgobj.add(line.svgobj)

        width = self.width/2
        line = svgbase.Line((0, 0), (width, width),
                            anchor_object=self.rect,
                            anchor_pos='ileft:ibottom',
                            **kwargs)
        self.svgobj.add(line.svgobj)

        line = svgbase.Line((0, 0), (width, width),
                            anchor_object=self.rect,
                            anchor_pos='iright:ibottom',
                            **kwargs)
        self.svgobj.add(line.svgobj)




# -*- coding: utf-8 -*-
"""
    simulate.sequence
    ~~~~~~~~~~~~~~~~~

    DNA sequence manipulation. 

    @Copyright: (c) 2017 by Lingxi Chen (chanlingxi@gmail.com). 
    @License: LICENSE_NAME, see LICENSE for more details.
"""
complements = {
    'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'N',
    'a':'t', 'c':'g', 'g':'c', 't':'a', 'n':'n',
    }
def reverse_complement(seq):
    return ''.join([complements[x] for x in seq[::-1]])



# -*- coding: utf-8 -*-
"""
    io
    ~~~~~~~~~~~~


    @Copyright: (c) 2018-06 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""
import os
import gzip
import csv
import vcf
import yaml
import pysam


class Record(object):

    fields = []
    sep = '\t'

    def __init__(self, stream_or_str=None, parent=None, args=None, sep='\t',
                 **kwargs):
        self.parent = parent
        self.sep = sep
        args = args or {}

        if parent:
            self.from_parent()
        elif stream_or_str:
            self.parse(stream_or_str)
        elif args:
            self.set(**args)
        else:
            self.set(**kwargs)

    def set(self, **kwargs):

        for key in kwargs:
            setattr(self, key, kwargs[key])

        for k in kwargs.keys():
            if k not in self.fields:
                self.fields.append(k)

    def from_parent(self, **kwargs):
        pass

    def parse(self, stream_or_str):
        # parse a stream_or_string to record object
        pass

    def _validate(self, x, func=None):
        if func in [None, int, float]:
            if not x:
                return x
            if x == '.':
                return None
        if func is str:
            if not x:
                return ''
        if func:
            return func(x)
        return x

    def _parse_list(self, line, sep=','):
        if isinstance(line, list):
            return line
        if not line or line == '.':
            return []
        return line.split(sep)

    def _format_list(self, x, sep=','):
        return sep.join(map(self._format_value, x))

    def _format_value(self, x):  # None, int, float, list
        if x == 0:
            return '0'
        if not x:
            return '.'   # notice not 0 is True
        if type(x) == list:
            return self._format_list(x)
        return str(x)

    def _parse_value(self, x):  # None, int, float, list
        if x == '.':
            return None
        splits = x.split(',')
        if splits:
            return map(self._parse_value, splits)
        return float(x)

    def __repr__(self):
        res_list = []
        for field in self.fields:
            try:
                x = self._format_value(getattr(self, field))
                res_list.append(x)
            except Exception:
                pass
        return self.sep.join(res_list)

    def __str__(self):
        return self.__repr__()

    def __eq__(self):
        pass


def safe_open(fn, mode):
    if not fn:
        return None
    if fn.endswith('gz'):
        return gzip.open(fn, mode)
    else:
        return open(fn, mode)


class Writer(object):

    def __init__(self, fn=None, record_cls=Record,
                 fields_prefix='#',
                 *args, **kwargs):
        self.fn = fn

        self.fields_prefix = fields_prefix
        self.sep = record_cls.sep
        self.fields = record_cls.fields
        self.write('', 'w')

    def write_headers(self, mode='a'):
        pass

    def write_fields(self, fields=None, mode='a'):
        if fields is None:
            fields = self.fields

        line = self.fields_prefix + self.sep.join(fields) + '\n'
        with safe_open(self.fn, mode) as f:
            f.write(line)
        return True

    def write_chunk(self, chunk, mode='a', dynamic_fields=False):
        with safe_open(self.fn, mode) as f:
            for i, record in enumerate(chunk):
                if i == 0 and dynamic_fields:
                    self.write_fields(fields=record.fields)
                f.write(str(record) + '\n')

    def write(self, content, mode='a'):
        with safe_open(self.fn, mode) as f:
            f.write(content)


class Reader(object):  # currently just csv, tsv file

    record_cls = Record

    def __init__(self,
                 in_data=None, in_fn=None,
                 record_cls=Record,
                 file_type='csv', sep=',', has_header=False,
                 sample=None):
        self.in_data = in_data
        self.in_fn = in_fn
        self.record_cls = record_cls
        self.fields = record_cls.fields
        self.file_type = file_type
        self.sep = sep
        self.has_header = has_header

        self.sample = sample
        if isinstance(self.in_fn, str) and self.in_fn.endswith('gz'):
            self.tbx = pysam.TabixFile(self.in_fn)

        if isinstance(self.in_fn, list) or isinstance(self.in_data, list):
            self.in_iter = self._multiple_data(in_datas=in_data, in_fns=in_fn)
        else:
            self.in_iter = self._single_data(in_data=in_data, in_fn=in_fn)

    def _read(self, in_fn):

        for line in safe_open(in_fn, 'rt'):
            if self.file_type != 'vcf' and line.startswith('#'):
                continue

            yield line

    def _single_data(self, in_data=None, in_fn=None):
        if in_data:
            return in_data
        if in_fn:
            if self.file_type == 'csv':
                in_iter = self._csv_fn(in_fn)
            elif self.file_type == 'vcf':
                in_iter = self._vcf_fn(in_fn)
            elif self.file_type == 'yaml':
                in_iter = self._yaml_fn(in_fn)
            else:
                in_iter = None
        else:
            in_iter = None
        return in_iter

    def _csv_fn(self, in_data):
        if self.has_header:
            in_iter = csv.DictReader(
                self._read(in_data),
                delimiter=self.sep, skipinitialspace=True)
        else:
            in_iter = csv.DictReader(
                self._read(in_data),
                fieldnames=self.fields,
                delimiter=self.sep, skipinitialspace=True)
        return in_iter

    def _vcf_fn(self, in_data):
        in_iter = vcf.Reader(self._read(in_data))
        return in_iter

    def _yaml_fn(self, in_fn):
        docs = yaml.load_all(safe_open(in_fn, 'r'))
        for doc in docs:
            if not doc:
                continue
            yield doc

    def _multiple_data(self, in_datas=None, in_fns=None):
        if in_datas:
            for in_data in in_datas:
                for x in self._single_data(in_data=in_data):
                    yield x
        elif in_fns:
            for in_fn in in_fns:
                for x in self._single_data(in_fn=in_fn):
                    yield os.path.split(in_fn)[1], x

    def __iter__(self):
        return self.read_record()

    def conditional_read(self, condition_func=None):
        for meta, record in self:
            if condition_func and condition_func(record):
                yield meta, record

    def fetch(self, chrom, start, end, condition_func=None):
        for line in self.tbx.fetch(chrom, start, end):
            row = dict(zip(self.record_cls.fields, line.split(self.sep)))
            record = self.record_cls(args=row)
            if condition_func is None:
                yield None, record
            elif condition_func and condition_func(record):
                yield None, record

    def read_record(self):
        for i, row_or_record in enumerate(self.in_iter):
            if isinstance(row_or_record, dict):
                row_or_record['sample'] = self.sample
                record = self.record_cls(args=row_or_record)
                meta = None

            elif isinstance(row_or_record, tuple):
                meta, record = row_or_record
                if isinstance(record, dict):
                    record['sample'] = self.sample
                    record = self.record_cls(args=record)
                elif not isinstance(record, self.record_cls):
                    record.sample = self.sample
                    record = self.record_cls(parent=record)
            elif not isinstance(row_or_record, self.record_cls):
                meta = None
                row_or_record.sample = self.sample
                record = self.record_cls(parent=row_or_record)
            else:
                meta = None
                record.sample = self.sample
                record = row_or_record

            yield meta, record

    def read_chunks(self, field):
        chunks = []
        prev_value = None

        for i, record in enumerate(self.read_record()):
            meta, record = record

            value = getattr(record, field)

            if prev_value is None or prev_value == value:
                chunks.append(record)
            else:  # update
                if i:
                    yield prev_value, chunks
                chunks = [record]
            prev_value = value

        if chunks:
            yield prev_value, chunks



# -*- coding: utf-8 -*-
"""
    basic
    ~~~~~~~~~~~~~~~

    basic operation on linux

    @Copyright: (c) 2018-04 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""

import gzip


def safe_open(fn, mod):
    if fn.endswith('gz'):
        return gzip.open(fn, mod)
    else:
        return open(fn, mod)
